from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion

import mayan.apps.databases.model_mixins
import mayan.apps.dynamic_search.model_mixins


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('dynamic_search', '0003_auto_20161028_0707')
    ]

    operations = [
        migrations.CreateModel(
            name='SavedResultset',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'time_to_live', models.PositiveIntegerField(
                        help_text='Time to keep the resultset in seconds. '
                        'This value is increased every time the resultset is '
                        'accessed.', verbose_name='Time to live'
                    )
                ),
                (
                    'timestamp', models.DateTimeField(
                        db_index=True, help_text='The server date and time '
                        'when the resultset was created.',
                        verbose_name='Timestamp'
                    )
                ),
                (
                    'epoch', models.PositiveBigIntegerField(
                        verbose_name='Epoch'
                    )
                ),
                (
                    'app_label', models.CharField(
                        max_length=64, verbose_name='App label'
                    )
                ),
                (
                    'model_name', models.CharField(
                        max_length=64, verbose_name='Model name'
                    )
                ),
                (
                    'search_query_text', models.TextField(
                        verbose_name='Search query'
                    )
                ),
                (
                    'search_explainer_text', models.TextField(
                        verbose_name='Search explainer text'
                    )
                ),
                (
                    'result_count', models.PositiveIntegerField(
                        help_text='Number of results stored in the '
                        'resultset.', verbose_name='Result count'
                    )
                ),
                (
                    'user', models.ForeignKey(
                        help_text='User for which the resultset was created.',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='saved_resultsets',
                        to=settings.AUTH_USER_MODEL, verbose_name='User'
                    )
                ),
            ],
            options={
                'verbose_name': 'Saved resultset',
                'verbose_name_plural': 'Saved resultsets',
                'ordering': ('-timestamp', 'user')
            },
            bases=(
                mayan.apps.databases.model_mixins.ExtraDataModelMixin,
                mayan.apps.dynamic_search.model_mixins.SavedResultsetBusinessLogicModelMixin,
                models.Model
            )
        ),
        migrations.CreateModel(
            name='SavedResultsetEntry',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'object_id', models.BigIntegerField()
                ),
                (
                    'saved_resultset', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='entries',
                        to='dynamic_search.savedresultset',
                        verbose_name='Saved resultset'
                    )
                )
            ],
            options={
                'verbose_name': 'Saved resultset entry',
                'verbose_name_plural': 'Saved resultset entries',
                'ordering': ('id',)
            }
        )
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('django_gpg', '0006_auto_20160510_0025')
    ]

    operations = [
        migrations.AlterField(
            model_name='key',
            name='creation_date',
            field=models.DateTimeField(
                editable=False, verbose_name='Creation date'
            ),
        ),
        migrations.AlterField(
            model_name='key',
            name='expiration_date',
            field=models.DateTimeField(
                blank=True, editable=False, null=True,
                verbose_name='Expiration date'
            ),
        )
    ]



from mayan.apps.source_compressed.source_backend_actions.mixins import (
    SourceBackendActionMixinCompressedInteractiveNot
)
from mayan.apps.source_stored_files.source_backend_actions.mixins import (
    SourceBackendActionMixinFileStoredInteractiveNot
)
from mayan.apps.sources.permissions import permission_sources_edit
from mayan.apps.sources.source_backend_actions.base import SourceBackendAction
from mayan.apps.sources.source_backend_actions.mixins.callback_mixins import (
    SourceBackendActionMixinCallbackDocumentUpload
)
from mayan.apps.sources.source_backend_actions.mixins.document_language_mixins import (
    SourceBackendActionMixinLanguageInteractiveNot
)
from mayan.apps.sources.source_backend_actions.mixins.document_mixins import (
    SourceBackendActionMixinDocumentUploadInteractiveNot
)
from mayan.apps.sources.source_backend_actions.mixins.document_type_mixins import (
    SourceBackendActionMixinDocumentTypeInteractiveNot
)
from mayan.apps.sources.source_backend_actions.mixins.user_mixins import (
    SourceBackendActionMixinUserInteractiveNot
)

from .mixins import SourceBackendActionMixinPeriodic


class SourceBackendActionPeriodicDocumentUpload(
    SourceBackendActionMixinDocumentUploadInteractiveNot,
    SourceBackendActionMixinDocumentTypeInteractiveNot,
    SourceBackendActionMixinLanguageInteractiveNot,
    SourceBackendActionMixinCompressedInteractiveNot,
    SourceBackendActionMixinFileStoredInteractiveNot,
    SourceBackendActionMixinUserInteractiveNot,
    SourceBackendActionMixinPeriodic,
    SourceBackendActionMixinCallbackDocumentUpload, SourceBackendAction
):
    name = 'document_upload'
    permission = permission_sources_edit
    stored_file_identifier_name = 'encoded_filename'
    stored_method_name_file_cleanup = 'action_file_delete'



from mayan.apps.icons.icons import Icon

icon_quota_backend_selection = Icon(
    driver_name='fontawesome-dual', primary_symbol='tachometer-alt',
    secondary_symbol='plus'
)
icon_quota_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='tachometer-alt',
    secondary_symbol='plus'
)
icon_quota_delete = Icon(driver_name='fontawesome', symbol='times')
icon_quota_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')
icon_quota_list = Icon(driver_name='fontawesome', symbol='tachometer-alt')
icon_quota_setup = Icon(driver_name='fontawesome', symbol='tachometer-alt')



from mayan.apps.common.class_mixins import AppsModuleLoaderMixin


class DocumentCreateWizardStep(AppsModuleLoaderMixin):
    _deregistry = {}
    _loader_module_name = 'wizard_steps'
    _registry = {}

    @classmethod
    def deregister(cls, step):
        cls._deregistry[step.name] = step

    @classmethod
    def deregister_all(cls):
        for step in cls.get_all():
            cls.deregister(step=step)

    @classmethod
    def done(cls, wizard):
        return {}

    @classmethod
    def get(cls, name):
        for step in cls.get_all():
            if name == step.name:
                return step

        raise KeyError(
            'Wizard step `{}` not found.'.format(name)
        )

    @classmethod
    def get_all(cls):
        step_list = [
            step for step in cls._registry.values() if step.name not in cls._deregistry
        ]

        return sorted(step_list, key=lambda entry: entry.number)

    @classmethod
    def get_choices(cls, attribute_name):
        return [
            (
                step.name, getattr(step, attribute_name)
            ) for step in cls.get_all()
        ]

    @classmethod
    def get_form_initial(cls, wizard):
        return {}

    @classmethod
    def get_form_kwargs(cls, wizard):
        return {}

    @classmethod
    def get_next_is_enabled(cls, wizard):
        return True

    @classmethod
    def post_upload_process(
        cls, document, query_string, source_id, user_id
    ):
        for step in cls.get_all():
            step.step_post_upload_process(
                document=document, query_string=query_string,
                source_id=source_id, user_id=user_id
            )

    @classmethod
    def register(cls, step):
        if step.name in cls._registry:
            raise Exception(
                'A step with this name already exists: %s' % step.name
            )

        if step.number in [reigstered_step.number for reigstered_step in cls.get_all()]:
            raise Exception(
                'A step with this number already exists: %s' % step.name
            )

        cls._registry[step.name] = step

    @classmethod
    def reregister(cls, name):
        cls._deregistry.pop(name)

    @classmethod
    def reregister_all(cls):
        cls._deregistry = {}

    @classmethod
    def step_post_upload_process(
        cls, document, query_string, source_id, user_id
    ):
        """
        Optional method executed when the wizard ends to allow the step to
        perform its action.
        """



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Documents'), name='documents'
)

# Document

event_document_created = namespace.add_event_type(
    label=_(message='Document created'), name='document_create'
)
event_document_edited = namespace.add_event_type(
    label=_(message='Document edited'), name='document_edit'
)
event_document_viewed = namespace.add_event_type(
    label=_(message='Document viewed'), name='document_view'
)

# Document File

event_document_file_created = namespace.add_event_type(
    label=_(message='Document file created'), name='document_file_created'
)
event_document_file_deleted = namespace.add_event_type(
    label=_(message='Document file deleted'), name='document_file_deleted'
)
event_document_file_edited = namespace.add_event_type(
    label=_(message='Document file edited'), name='document_file_edited'
)

# Document type

event_document_type_created = namespace.add_event_type(
    label=_(message='Document type created'), name='document_type_created'
)
event_document_type_edited = namespace.add_event_type(
    label=_(message='Document type edited'), name='document_type_edit'
)
event_document_type_quick_label_created = namespace.add_event_type(
    label=_(message='Document type quick label created'),
    name='document_type_quick_label_created'
)
event_document_type_quick_label_deleted = namespace.add_event_type(
    label=_(message='Document type quick label deleted'),
    name='document_type_quick_label_deleted'
)
event_document_type_quick_label_edited = namespace.add_event_type(
    label=_(message='Document type quick label edited'),
    name='document_type_quick_label_edited'
)
# The type of an existing document is changed to another type.
event_document_type_changed = namespace.add_event_type(
    label=_(message='Document type changed'), name='document_type_change'
)

# Document version

event_document_version_created = namespace.add_event_type(
    label=_(message='Document version created'), name='document_version_created'
)
event_document_version_deleted = namespace.add_event_type(
    label=_(message='Document version deleted'), name='document_version_deleted'
)
event_document_version_edited = namespace.add_event_type(
    label=_(message='Document version edited'), name='document_version_edited'
)

# Document version page

event_document_version_page_created = namespace.add_event_type(
    label=_(message='Document version page created'),
    name='document_version_page_created'
)
event_document_version_page_deleted = namespace.add_event_type(
    label=_(message='Document version page deleted'),
    name='document_version_page_deleted'
)
event_document_version_page_edited = namespace.add_event_type(
    label=_(message='Document version page edited'),
    name='document_version_page_edited'
)

# Trashed document

event_document_trashed = namespace.add_event_type(
    label=_(message='Document trashed'), name='document_trashed'
)
event_trashed_document_deleted = namespace.add_event_type(
    label=_(message='Trashed document deleted'), name='trashed_document_deleted'
)
event_trashed_document_restored = namespace.add_event_type(
    label=_(message='Trashed document restored'), name='trashed_document_restored'
)

# Historic events

event_document_download = namespace.add_event_type(
    label=_(message='Document downloaded (historic)'), name='document_download'
)
event_document_file_pre_save = namespace.add_event_type(
    label=_(message='Document file created (historic)'),
    name='document_file_pre_save'
)
event_document_version_revert = namespace.add_event_type(
    label=_(message='Document version reverted (historic)'),
    name='document_version_revert'
)
event_document_version_new = namespace.add_event_type(
    label=_(message='Document version uploaded (historic)'),
    name='document_new_version'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link
from mayan.apps.navigation.utils import factory_condition_queryset_access

from ..icons import (
    icon_document_type_create, icon_document_type_delete,
    icon_document_type_edit, icon_document_type_filename_create,
    icon_document_type_filename_delete, icon_document_type_filename_edit,
    icon_document_type_filename_generator, icon_document_type_filename_list,
    icon_document_type_list, icon_document_type_retention_policies,
    icon_document_type_setup
)
from ..permissions import (
    permission_document_type_create, permission_document_type_delete,
    permission_document_type_edit, permission_document_type_view
)

link_document_type_create = Link(
    icon=icon_document_type_create,
    permission=permission_document_type_create,
    text=_(message='Create document type'), view='documents:document_type_create'
)
link_document_type_delete = Link(
    args='resolved_object.id', icon=icon_document_type_delete,
    permission=permission_document_type_delete, tags='dangerous',
    text=_(message='Delete'), view='documents:document_type_delete'
)
link_document_type_edit = Link(
    args='resolved_object.id', icon=icon_document_type_edit,
    permission=permission_document_type_edit, text=_(message='Edit'),
    view='documents:document_type_edit'
)
link_document_type_filename_create = Link(
    args='document_type.id',
    icon=icon_document_type_filename_create,
    permission=permission_document_type_edit,
    text=_(message='Add quick label to document type'),
    view='documents:document_type_filename_create'
)
link_document_type_filename_delete = Link(
    args='resolved_object.id',
    icon=icon_document_type_filename_delete,
    permission=permission_document_type_edit,
    tags='dangerous', text=_(message='Delete'),
    view='documents:document_type_filename_delete'
)
link_document_type_filename_edit = Link(
    args='resolved_object.id',
    icon=icon_document_type_filename_edit,
    permission=permission_document_type_edit,
    text=_(message='Edit'), view='documents:document_type_filename_edit'
)
link_document_type_filename_list = Link(
    args='resolved_object.id',
    icon=icon_document_type_filename_list,
    permission=permission_document_type_view,
    text=_(message='Quick labels'), view='documents:document_type_filename_list'
)
link_document_type_filename_generator = Link(
    args='resolved_object.id', icon=icon_document_type_filename_generator,
    permission=permission_document_type_edit,
    text=_(message='Filename generation'),
    view='documents:document_type_filename_generator'
)
link_document_type_list = Link(
    icon=icon_document_type_list, text=_(message='Document types'),
    view='documents:document_type_list'
)
link_document_type_retention_policies = Link(
    args='resolved_object.id',
    icon=icon_document_type_retention_policies,
    permission=permission_document_type_edit,
    text=_(message='Retention policies'),
    view='documents:document_type_retention_policies'
)
link_document_type_setup = Link(
    condition=factory_condition_queryset_access(
        app_label='documents', model_name='DocumentType',
        object_permission=permission_document_type_view,
        view_permission=permission_document_type_create,
    ), icon=icon_document_type_setup, text=_(message='Document types'),
    view='documents:document_type_list'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.serializers.document_file_serializers import (
    DocumentFileSerializer
)
from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import MultiKwargHyperlinkedIdentityField

from ..models import DocumentFileDriverEntry, FileMetadataEntry

from .file_metadata_serializers import StoredDriverSerializer


class DocumentFileMetadataDriverEntrySerializer(
    serializers.HyperlinkedModelSerializer
):
    document_file = DocumentFileSerializer(
        label=_(message='Document file'), read_only=True
    )
    entries_url = MultiKwargHyperlinkedIdentityField(
        label=_('URL'), view_kwargs=(
            {
                'lookup_field': 'document_file.document_id',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'document_file_id',
                'lookup_url_kwarg': 'document_file_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'driver_id'
            }
        ), view_name='rest_api:document_file_metadata_entry-list'
    )
    stored_driver = StoredDriverSerializer(
        label=_(message='Stored driver'), read_only=True, source='driver'
    )
    url = MultiKwargHyperlinkedIdentityField(
        label=_('URL'), view_kwargs=(
            {
                'lookup_field': 'document_file.document_id',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'document_file_id',
                'lookup_url_kwarg': 'document_file_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'driver_id'
            }
        ), view_name='rest_api:document_file_metadata_driver-detail'
    )

    class Meta:
        fields = (
            'document_file', 'entries_url', 'id', 'stored_driver', 'url'
        )
        model = DocumentFileDriverEntry
        read_only_fields = (
            'document_file', 'entries_url', 'id', 'stored_driver', 'url'
        )


class DocumentFileMetadataEntrySerializer(
    serializers.HyperlinkedModelSerializer
):
    url = MultiKwargHyperlinkedIdentityField(
        label=_('URL'), view_kwargs=(
            {
                'lookup_field': 'document_file_driver_entry.document_file.document_id',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'document_file_driver_entry.document_file_id',
                'lookup_url_kwarg': 'document_file_id'
            },
            {
                'lookup_field': 'document_file_driver_entry.pk',
                'lookup_url_kwarg': 'driver_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'entry_id'
            }
        ), view_name='rest_api:document_file_metadata_entry-detail'
    )

    class Meta:
        fields = ('id', 'internal_name', 'key', 'url', 'value')
        model = FileMetadataEntry
        read_only_fields = ('id', 'internal_name', 'key', 'url', 'value')



from django.db import migrations


def code_document_type_driver_configuration_populate(apps, schema_editor):
    DocumentType = apps.get_model(
        app_label='documents', model_name='DocumentType'
    )
    DocumentTypeDriverConfiguration = apps.get_model(
        app_label='file_metadata', model_name='DocumentTypeDriverConfiguration'
    )
    DocumentTypeSettings = apps.get_model(
        app_label='file_metadata', model_name='DocumentTypeSettings'
    )
    StoredDriver = apps.get_model(
        app_label='file_metadata', model_name='StoredDriver'
    )

    for document_type in DocumentType.objects.using(alias=schema_editor.connection.alias).all():
        try:
            document_type_file_metadata_settings = DocumentTypeSettings.objects.get(document_type=document_type)
        except DocumentTypeSettings.DoesNotExist:
            enabled = True
        else:
            enabled = document_type_file_metadata_settings.auto_process

        for stored_driver in StoredDriver.objects.using(alias=schema_editor.connection.alias).all():
            # TODO: Add `create_defaults={'enabled': enabled}` when Django
            # is updated to 5.0.
            DocumentTypeDriverConfiguration.objects.update_or_create(
                defaults={'enabled': enabled}, document_type=document_type,
                stored_driver=stored_driver
            )


class Migration(migrations.Migration):
    dependencies = [
        ('file_metadata', '0013_documenttypedriverconfiguration')
    ]

    operations = [
        migrations.RunPython(
            code=code_document_type_driver_configuration_populate,
            reverse_code=migrations.RunPython.noop
        )
    ]



from .runtime import cache


def handler_document_cache_delete(sender, **kwargs):
    cache.clear_document(
        document=kwargs['instance']
    )


def handler_node_cache_delete(sender, **kwargs):
    cache.clear_node(
        node=kwargs['instance']
    )



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('common', '0007_auto_20170118_1758')
    ]

    operations = [
        migrations.CreateModel(
            name='ErrorLogEntry',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'namespace', models.CharField(
                        max_length=128, verbose_name='Namespace'
                    )
                ),
                (
                    'object_id', models.PositiveIntegerField(
                        blank=True, null=True
                    )
                ),
                (
                    'datetime', models.DateTimeField(
                        auto_now_add=True, db_index=True,
                        verbose_name='Date time'
                    )
                ),
                (
                    'result', models.TextField(
                        blank=True, null=True, verbose_name='Result'
                    )
                ),
                (
                    'content_type', models.ForeignKey(
                        blank=True, null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='error_log_content_type',
                        to='contenttypes.ContentType'
                    )
                )
            ],
            options={
                'ordering': ('datetime',),
                'verbose_name': 'Error log entry',
                'verbose_name_plural': 'Error log entries'
            }
        )
    ]



import uuid

from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.backends.classes import BaseBackend

__all__ = (
    'BaseDocumentFilenameGenerator', 'OriginalDocumentFilenameGenerator',
    'UUIDDocumentFilenameGenerator'
)


class BaseDocumentFilenameGenerator:
    default = None
    _registry = {}

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    @classmethod
    def get_choices(cls):
        choices = []

        for name, klass in cls._registry.items():
            choices.append(
                (
                    name, format_lazy(
                        '{} - {}', klass.label, klass.description
                    )
                )
            )

        return sorted(choices)

    @classmethod
    def get_default(cls):
        for backend in cls._registry.values():
            if backend.default:
                return backend.name

    @classmethod
    def register(cls, klass):
        cls._registry[klass.name] = klass

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def upload_to(self, instance, filename):
        raise NotImplementedError


class DocumentFileAction(BaseBackend):
    _backend_identifier = 'action_id'
    _loader_module_name = 'document_file_actions'


class DocumentVersionModification(BaseBackend):
    _loader_module_name = 'document_version_modifications'


class OriginalDocumentFilenameGenerator(BaseDocumentFilenameGenerator):
    name = 'original'
    label = _(message='Original')
    description = _(
        message='Keeps the original filename of the uploaded file.'
    )

    def upload_to(self, instance, filename):
        return '{}'.format(instance.document.label)


class UUIDDocumentFilenameGenerator(BaseDocumentFilenameGenerator):
    default = True
    name = 'uuid'
    label = _(message='UUID')
    description = _(
        message='Generates an immutable, random UUID (RFC 4122) for each file.'
    )

    def upload_to(self, instance, filename):
        return str(
            uuid.uuid4()
        )


class UUIDPlusOriginalFilename(BaseDocumentFilenameGenerator):
    name = 'uuid_plus_original'
    label = _(message='UUID plus original')
    description = _(
        message='Generates an immutable, random UUID (RFC 4122) for each file and '
        'appends the original filename of the uploaded file.'
    )

    def upload_to(self, instance, filename):
        return '{}_{}'.format(
            uuid.uuid4(), instance.document.label
        )


BaseDocumentFilenameGenerator.register(
    klass=OriginalDocumentFilenameGenerator
)
BaseDocumentFilenameGenerator.register(klass=UUIDDocumentFilenameGenerator)
BaseDocumentFilenameGenerator.register(klass=UUIDPlusOriginalFilename)



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import (
    DEFAULT_SIGNATURES_STORAGE_BACKEND,
    DEFAULT_SIGNATURES_STORAGE_BACKEND_ARGUMENTS
)
from .setting_migrations import DocumentSignaturesSettingMigration

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Document signatures'),
    migration_class=DocumentSignaturesSettingMigration, name='signatures',
    version='0002'
)

setting_storage_backend = setting_namespace.do_setting_add(
    default=DEFAULT_SIGNATURES_STORAGE_BACKEND,
    global_name='SIGNATURES_STORAGE_BACKEND', help_text=_(
        message='Path to the Storage subclass to use when storing detached '
        'signatures.'
    )
)
setting_storage_backend_arguments = setting_namespace.do_setting_add(
    default=DEFAULT_SIGNATURES_STORAGE_BACKEND_ARGUMENTS,
    global_name='SIGNATURES_STORAGE_BACKEND_ARGUMENTS', help_text=_(
        message='Arguments to pass to the SIGNATURE_STORAGE_BACKEND.'
    )
)



from mayan.apps.dependencies.classes import PythonDependency
from mayan.apps.dependencies.environments import environment_documentation

PythonDependency(
    environment=environment_documentation, module=__name__,
    name='Sphinx', version_string='==8.2.3'
)
PythonDependency(
    environment=environment_documentation, module=__name__,
    name='sphinx-sitemap', version_string='==2.8.0'
)
PythonDependency(
    environment=environment_documentation, module=__name__,
    name='sphinx_rtd_theme', version_string='==3.0.2'
)
PythonDependency(
    environment=environment_documentation, module=__name__,
    name='sphinxcontrib-spelling', version_string='==8.0.1'
)



from django.core.exceptions import FieldDoesNotExist
from django.db import migrations, models
from django.db.utils import IntegrityError, OperationalError


def code_do_unique_on_columns_drop(apps, schema_editor):
    Model = apps.get_model(
        app_label='sources', model_name='DocumentFileSourceMetadata'
    )

    connection = schema_editor.connection
    database_columns = ('document_file_id', 'key')
    field_names = ('document_file', 'key'),
    table_name = Model._meta.db_table
    target = set(database_columns)

    with connection.cursor() as cursor:
        constraints = connection.introspection.get_constraints(
            cursor=cursor, table_name=table_name
        )

        for constraint_name, constraint_data in constraints.items():
            constraint_columns = set(
                constraint_data.get('columns') or []
            )

            if constraint_data.get('unique') and constraint_columns == target:
                # Try dropping as a constraint first.
                try:
                    schema_editor.remove_constraint(
                        model=Model,
                        constraint=models.UniqueConstraint(
                            fields=field_names, name=constraint_name
                        )
                    )
                except IntegrityError:
                    # The uniqueness is not coded as a constraint.
                    # Fall back to dropping as an index.
                    try:
                        schema_editor.remove_index(
                            model=Model,
                            index=models.Index(
                                fields=field_names, name=constraint_name
                            )
                        )
                    except OperationalError:
                        """It's neither or deleted already."""
                else:
                    continue


def do_source_column_remove_if_exists(apps, schema_editor):
    Model = apps.get_model(
        app_label='sources', model_name='DocumentFileSourceMetadata'
    )

    connection = schema_editor.connection
    source_field_name = 'source'
    source_column_name = 'source_id'
    table = Model._meta.db_table

    with connection.cursor() as cursor:
        table_field_list = connection.introspection.get_table_description(
            cursor=cursor, table_name=table
        )

        columns_set = {
            field.name for field in table_field_list
        }

        if source_column_name not in columns_set:
            # Field is already gone, exit.
            return
        else:
            try:
                field = Model._meta.get_field(field_name=source_field_name)
            except FieldDoesNotExist:
                # Add a fake field.
                field = models.IntegerField(null=True)
                field.set_attributes_from_name(name=source_field_name)
                field.db_column = source_column_name

            # Use backend DDL to DROP COLUMN safely across vendors.
            schema_editor.remove_field(model=Model, field=field)


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0090_alter_documentversion_active'),
        ('sources', '0031_alter_documentfilesourcemetadata_value')
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            database_operations=[
                migrations.RunPython(
                    code=code_do_unique_on_columns_drop,
                    reverse_code=migrations.RunPython.noop
                )
            ],
            state_operations=[
                migrations.AlterUniqueTogether(
                    name="documentfilesourcemetadata",
                    unique_together=set()
                )
            ]
        ),
        migrations.AlterUniqueTogether(
            name='documentfilesourcemetadata', unique_together={
                ('document_file', 'key')
            }
        ),
        migrations.SeparateDatabaseAndState(
            database_operations=[
                migrations.RunPython(
                    code=do_source_column_remove_if_exists,
                    reverse_code=migrations.RunPython.noop
                )
            ],
            state_operations=[
                migrations.RemoveField(
                    model_name='documentfilesourcemetadata',
                    name='source'
                )
            ]
        )
    ]



from django.urls import re_path

from .views import ModelPropertyListView, PropertyModelListView

urlpatterns = [
    re_path(
        route=r'^models/$', name='property_model_list',
        view=PropertyModelListView.as_view()
    ),
    re_path(
        route=r'^apps/(?P<app_label>[-\w]+)/models/(?P<model_name>[-\w]+)/properties/$',
        name='model_property_list', view=ModelPropertyListView.as_view()
    ),


]



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import (
    icon_quota_create, icon_quota_delete, icon_quota_edit, icon_quota_list,
    icon_quota_setup
)
from .permissions import (
    permission_quota_create, permission_quota_delete, permission_quota_edit,
    permission_quota_view
)

link_quota_create = Link(
    icon=icon_quota_create, permission=permission_quota_create,
    text=_(message='Create quota'), view='quotas:quota_backend_selection'
)
link_quota_delete = Link(
    args='resolved_object.pk', icon=icon_quota_delete,
    permission=permission_quota_delete, tags='dangerous',
    text=_(message='Delete'), view='quotas:quota_delete'
)
link_quota_edit = Link(
    args='object.pk', icon=icon_quota_edit,
    permission=permission_quota_edit, text=_(message='Edit'),
    view='quotas:quota_edit'
)
link_quota_list = Link(
    icon=icon_quota_list, text=_(message='Quotas list'), view='quotas:quota_list'
)
link_quota_setup = Link(
    condition=factory_condition_queryset_access(
        app_label='quotas', model_name='Quota',
        object_permission=permission_quota_view,
        view_permission=permission_quota_create,
    ), icon=icon_quota_setup, text=_(message='Quotas'),
    view='quotas:quota_list'
)



from mayan.apps.documents.api_views.api_view_mixins import (
    ParentObjectDocumentAPIViewMixin
)
from mayan.apps.storage.api_views.base import APIObjectDownloadView
from mayan.apps.storage.views.mixins import ViewMixinBackendDownload

from .events import event_document_file_downloaded
from .permissions import permission_document_file_download
from .settings import (
    setting_document_file_download_backend,
    setting_document_file_download_backend_arguments
)


class APIDocumentFileDownloadView(
    ViewMixinBackendDownload, ParentObjectDocumentAPIViewMixin,
    APIObjectDownloadView
):
    """
    get: Download a document file.
    """
    backend_arguments = setting_document_file_download_backend_arguments.value
    backend_path = setting_document_file_download_backend.value
    download_event_type = event_document_file_downloaded
    lookup_url_kwarg = 'document_file_id'
    mayan_object_permission_map = {'GET': permission_document_file_download}

    def get_download_filename(self):
        return self.get_object().filename

    def get_serializer(self, *args, **kwargs):
        return None

    def get_serializer_class(self):
        return None

    def get_source_queryset(self):
        return self.get_document().files.all()

    def retrieve(self, request, *args, **kwargs):
        return self.render_to_response()



from mayan.apps.rest_api import generics

from ..views.mixins import ViewMixinDownload, ViewMixinDownloadEvent


class APIObjectDownloadView(
    ViewMixinDownloadEvent, ViewMixinDownload, generics.RetrieveAPIView
):
    def get_download_event_target(self):
        return self.get_object()

    def get_download_file_object(self):
        instance = self.get_object()
        return instance.open(mode='rb')

    def get_serializer(self, *args, **kwargs):
        return None

    def get_serializer_class(self):
        return None

    def retrieve(self, request, *args, **kwargs):
        return self.render_to_response()


class APIObjectDownloadBackendDownloadView(
    ViewMixinDownloadEvent, ViewMixinDownload, generics.RetrieveAPIView
):
    def get_download_event_target(self):
        return self.get_object()

    def get_download_file_object(self):
        instance = self.get_object()
        return instance.open(mode='rb')

    def get_serializer(self, *args, **kwargs):
        return None

    def get_serializer_class(self):
        return None

    def retrieve(self, request, *args, **kwargs):
        return self.render_to_response()



import logging

from django.apps import apps
from django.conf import settings
from django.db.utils import IntegrityError, OperationalError, ProgrammingError
from django.utils.functional import classproperty
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin
from mayan.apps.common.utils import (
    convert_to_internal_name, deduplicate_dictionary_values,
    get_class_full_name
)
from mayan.apps.templating.template_backends import Template

from .exceptions import FileMetadataError
from .literals import ERROR_LOG_DOMAIN_NAME
from .settings import setting_auto_process, setting_drivers_arguments

logger = logging.getLogger(name=__name__)


class FileMetadataDriverCollection:
    _driver_to_mime_type_dict = {}
    _mime_type_to_driver_dict = {}

    @classmethod
    def do_driver_register(cls, klass):
        if klass in cls._driver_to_mime_type_dict:
            raise FileMetadataError(
                'Driver "{}" is already registered.'.format(klass)
            )

        cls._driver_to_mime_type_dict[klass] = klass.mime_type_list

        for mime_type in klass.mime_type_list:
            cls._mime_type_to_driver_dict.setdefault(
                mime_type, []
            ).append(klass)

        klass.dotted_path = get_class_full_name(klass=klass)

    @classmethod
    def get_all(cls, sorted=False):
        result = list(
            cls._driver_to_mime_type_dict.keys()
        )
        if sorted:
            result.sort(key=lambda driver: driver.label)

        return result

    @classmethod
    def get_driver_for_mime_type(cls, mime_type):
        driver_class_list = cls._mime_type_to_driver_dict.get(
            mime_type, ()
        )
        # Add wildcard drivers, drivers meant to be executed for all MIME
        # types.
        driver_class_list += tuple(
            cls._mime_type_to_driver_dict.get(
                '*', ()
            )
        )
        driver_class_list = [
            driver_class for driver_class in driver_class_list if driver_class.model_instance
        ]

        return driver_class_list


class FileMetadataDriverMetaclass(type):
    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(mcs, name, bases, attrs)

        if not new_class.__module__ == __name__:
            FileMetadataDriverCollection.do_driver_register(klass=new_class)

        return new_class


class FileMetadataDriver(
    AppsModuleLoaderMixin, metaclass=FileMetadataDriverMetaclass
):
    _loader_module_name = 'drivers'
    argument_name_list = ()
    description = ''
    # Workaround a Django bug that causes the template system to call the
    # class which would cause it to create an instance without required
    # arguments and lead to an empty entry in list views.
    # https://stackoverflow.com/questions/6861601/cannot-resolve-callable-context-variable
    do_not_call_in_templates = True
    dotted_path_previous_list = ()
    enabled = None
    """
    Default value the `enabled` field when driver is populated or new
    document types are added.
    `None` the final value will be up to the setting `setting_auto_process`.
    `True` will be enabled.
    `False` will not be enabled.
    Don't access the property directly, use cls.get_enabled_value().
    """
    internal_name = None
    label = None
    mime_type_list = ()
    model_instance = None

    @classproperty
    def collection(cls):
        if cls != FileMetadataDriver:
            raise AttributeError(
                'This method is only available to the parent class.'
            )
        return FileMetadataDriverCollection

    @classmethod
    def do_document_file_process(cls, document_file):
        arguments = cls.get_argument_values_for_document_file(
            document_file=document_file
        )

        driver_instance = cls(**arguments)

        driver_instance._instance_do_document_file_process(
            document_file=document_file
        )

    @classmethod
    def do_model_instance_populate(cls):
        DocumentType = apps.get_model(
            app_label='documents', model_name='DocumentType'
        )
        DocumentTypeDriverConfiguration = apps.get_model(
            app_label='file_metadata',
            model_name='DocumentTypeDriverConfiguration'
        )
        StoredDriver = apps.get_model(
            app_label='file_metadata', model_name='StoredDriver'
        )

        try:
            model_instance, created = StoredDriver.objects.update_or_create(
                driver_path=cls.dotted_path, defaults={
                    'exists': True, 'internal_name': cls.internal_name
                }
            )
        except IntegrityError:
            # May be a driver that moved to another location.
            model_instance = StoredDriver.objects.get(
                internal_name=cls.internal_name
            )

            if model_instance.driver_path in cls.dotted_path_previous_list:
                model_instance.driver_path = cls.dotted_path
                model_instance.exists = True
                model_instance.save()

                created = False
            else:
                # Unknown situation, re-raise original error.
                raise

        cls.model_instance = model_instance

        if created:
            enabled = cls.get_enabled_value()
            for document_type in DocumentType.objects.all():
                DocumentTypeDriverConfiguration.objects.update_or_create(
                    defaults={'enabled': enabled},
                    document_type=document_type, stored_driver=model_instance
                )

    @classmethod
    def get_argument_name_list(cls):
        return cls.argument_name_list

    @classmethod
    def get_argument_values_for_document_file(cls, document_file):
        document_type = document_file.document.document_type

        configuration_instance = document_type.file_metadata_driver_configurations.get(
            stored_driver__internal_name=cls.internal_name
        )

        document_type_arguments = configuration_instance.get_arguments()

        context = {'document_file': document_file}

        argument_values_rendered = {}

        for key, value in document_type_arguments.items():
            template = Template(template_string=value)
            template_result = template.render(context=context)
            argument_values_rendered[key] = template_result

        return argument_values_rendered

    @classmethod
    def get_argument_values_from_settings(cls):
        result = {}

        raw_argument_values = setting_drivers_arguments.value.get(
            cls.internal_name, {}
        )

        argument_name_list = cls.get_argument_name_list()

        for argument_name in argument_name_list:
            value = raw_argument_values.get(argument_name)
            if value:
                result[argument_name] = value

        return result

    @classmethod
    def get_argument_values_from_settings_display(cls):
        try:
            arguments = cls.get_argument_values_from_settings()
        except Exception as exception:
            arguments = {
                '__error__': _(
                    message='Badly formatted arguments YAML; %s'
                ) % exception
            }

        result = []

        argument_name_list = cls.get_argument_name_list()

        for argument_name in argument_name_list:
            value = arguments.get(argument_name)
            result.append(
                '{}: {}'.format(argument_name, value)
            )

        return ', '.join(result)

    @classmethod
    def get_enabled_value_from_settings(cls):
        raw_argument_values = setting_drivers_arguments.value.get(
            cls.internal_name, {}
        )

        return raw_argument_values.get('enabled', None)

    @classmethod
    def get_enabled_value(cls):
        """
        Calculate the final value of the class enabled field when populated
        or for new document types.

        - `None` the final value will be up to the setting
          `setting_auto_process`.
        - `True` will be enabled.
        - `False` will not be enabled.

        The value itself an be overridden by the setting
        `FILE_METADATA_DRIVERS_ARGUMENTS` to change the behavior per driver.
        """

        enabled_from_settings = cls.get_enabled_value_from_settings()

        if enabled_from_settings is None:
            if cls.enabled is None:
                return setting_auto_process.value
            else:
                return cls.enabled
        else:
            return enabled_from_settings

    @classmethod
    def get_form_class(cls):
        return getattr(cls, 'Form', None)

    @classmethod
    def get_mime_type_list_display(cls):
        return ', '.join(cls.mime_type_list)

    @classmethod
    def initialize(cls):
        StoredDriver = apps.get_model(
            app_label='file_metadata', model_name='StoredDriver'
        )

        try:
            """
            Check is the table is ready.
            If not, this will log an error similar to this:
            2023-12-12 09:12:46.985 UTC [79] ERROR:  relation "file_metadata_storeddriver" does not exist at character 22
            2023-12-12 09:12:46.985 UTC [79] STATEMENT:  SELECT 1 AS "a" FROM "file_metadata_storeddriver" LIMIT 1
            This error is expected and should be ignored.
            """
            StoredDriver.objects.exists()
        except (OperationalError, ProgrammingError):
            """
            This error is expected when trying to initialize the
            stored permissions during the initial creation of the
            database. Can be safely ignored under that situation.
            """
        else:
            try:
                # Reset all `StoredDriver` in case a file metadata app was
                # disabled.
                StoredDriver.objects.update(exists=False)
            except (OperationalError, ProgrammingError):
                """
                This error is expected when performing an upgrade between
                series 4.7 and 4.8 as the fields `exists` has not yet been
                created by the migration.
                """
            else:
                for driver in cls.collection.get_all():
                    driver.do_model_instance_populate()

    def _instance_do_document_file_process(self, document_file):
        try:
            logger.debug(
                'Starting processing document file: %s', document_file
            )

            FileMetadataEntry = apps.get_model(
                app_label='file_metadata', model_name='FileMetadataEntry'
            )

            file_metadata_dictionary = self._process(
                document_file=document_file
            )

            file_metadata_dictionary = file_metadata_dictionary or {}

            internal_name_dictionary = {}
            for key in file_metadata_dictionary.keys():
                internal_name_dictionary[key] = convert_to_internal_name(
                    value=key
                )

            internal_name_dictionary_deduplicated = deduplicate_dictionary_values(
                dictionary=internal_name_dictionary
            )

            queryset_document_file_metadata = self.model_instance.driver_entries.filter(
                document_file=document_file
            )
            queryset_document_file_metadata.delete()

            document_file_driver_entry = self.model_instance.driver_entries.create(
                document_file=document_file
            )

            coroutine = FileMetadataEntry.objects.create_bulk()
            next(coroutine)

            for key, value in file_metadata_dictionary.items():
                internal_name = internal_name_dictionary_deduplicated[key]

                # Drivers should not be returning `None` values.
                # Added to workaround undocumented backward incompatible
                # changes in Ollama.
                value_clean = value or ''

                coroutine.send(
                    {
                        'document_file_driver_entry': document_file_driver_entry,
                        'internal_name': internal_name, 'key': key,
                        'value': value_clean
                    }
                )

            coroutine.close()
        except Exception as exception:
            if settings.DEBUG and 0:
                raise
            else:
                error_log_text = '''
                Cannot process driver `{driver}` for document file `{document_file}`; {exception}
                '''.format(
                    document_file=document_file, driver=self.label,
                    exception=exception
                )
                document_file.error_log.create(
                    domain_name=ERROR_LOG_DOMAIN_NAME, text=error_log_text
                )

    def _process(self, document_file):
        raise NotImplementedError(
            'Your %s class has not defined the required _process() '
            'method.' % self.__class__.__name__
        )



import hashlib

from django.conf import settings
from django.core import serializers

from ..literals import (
    ERROR_LOG_DOMAIN_NAME, GRAPHVIZ_ID_STATE_ACTION,
    GRAPHVIZ_SYMBOL_CONDITIONAL, WORKFLOW_ACTION_WHEN_CHOICES
)


class WorkflowStateActionBusinessLogicMixin:
    def do_diagram_generate(self, diagram):
        if self.enabled:
            if self.has_condition():
                markup_state_action_label = '{} {}'.format(
                    self.label, GRAPHVIZ_SYMBOL_CONDITIONAL
                )
            else:
                markup_state_action_label = self.label

            markup_state_action_name = self.get_graph_id()
            markup_state_action_state = self.state.get_graph_id()

            node_kwargs = {
                'label': markup_state_action_label,
                'name': markup_state_action_name,
                'shape': 'box'
            }
            diagram.node(**node_kwargs)

            state_action_choice_dict = dict(WORKFLOW_ACTION_WHEN_CHOICES)
            edge_label = str(
                state_action_choice_dict[self.when]
            )
            edge_kwargs = {
                'arrowhead': 'dot',
                'arrowtail': 'dot',
                'dir': 'both',
                'label': edge_label,
                'head_name': markup_state_action_name,
                'style': 'dashed',
                'tail_name': markup_state_action_state
            }

            diagram.edge(**edge_kwargs)

    def get_graph_id(self):
        return '{}{}'.format(GRAPHVIZ_ID_STATE_ACTION, self.pk)

    def get_hash(self):
        return hashlib.sha256(
            string=serializers.serialize(
                format='json', queryset=(self,)
            ).encode()
        ).hexdigest()

    def execute(self, context, workflow_instance):
        condition_context = {'workflow_instance': workflow_instance}
        condition_result = self.evaluate_condition(context=condition_context)

        if condition_result:
            context.update(
                {'workflow_instance': workflow_instance}
            )

            try:
                backend_instance = self.get_backend_instance()
                backend_instance.execute(context=context)
            except Exception as exception:
                self.error_log.create(
                    domain_name=ERROR_LOG_DOMAIN_NAME,
                    text='{}; {}'.format(
                        exception.__class__.__name__, exception
                    )
                )

                if settings.DEBUG or settings.TESTING:
                    raise
            else:
                queryset_error_logs = self.error_log.filter(
                    domain_name=ERROR_LOG_DOMAIN_NAME
                )
                queryset_error_logs.delete()



from django.db import migrations


def code_document_version_active_populate(apps, schema_editor):
    cursor_main = schema_editor.connection.create_cursor(
        name='document_version_active'
    )
    cursor_document_version_active_update = schema_editor.connection.cursor()

    query_main = '''
        SELECT
            {documents_document}.{id}, {documents_documentversion_active}.{id}
        FROM (
            SELECT
                {documents_documentversion_active_timestamp_max}.{document_id}, {documents_documentversion}.{id}
            FROM {documents_documentversion}, (
                SELECT
                    {document_id}, MAX(timestamp) as {timestamp_max}
                FROM
                    {documents_documentversion}
                WHERE
                    {active}
                GROUP BY
                    {document_id}, {active}
            ) AS {documents_documentversion_active_timestamp_max}
            WHERE {documents_documentversion}.{timestamp} = {documents_documentversion_active_timestamp_max}.{timestamp_max}
            AND {documents_documentversion}.{document_id} = {documents_documentversion_active_timestamp_max}.{document_id}
        ) as {documents_documentversion_active}
        INNER JOIN
            {documents_document}
        ON
            {documents_document}.{id} = {documents_documentversion_active}.{document_id}
    '''.format(
        active=schema_editor.connection.ops.quote_name(
            name='active'
        ),
        documents_document=schema_editor.connection.ops.quote_name(
            name='documents_document'
        ),
        document_id=schema_editor.connection.ops.quote_name(
            name='document_id'
        ),
        id=schema_editor.connection.ops.quote_name(
            name='id'
        ),
        documents_documentversion=schema_editor.connection.ops.quote_name(
            name='documents_documentversion'
        ),
        documents_documentversion_active=schema_editor.connection.ops.quote_name(
            name='documents_documentversion_active'
        ),
        documents_documentversion_active_timestamp_max=schema_editor.connection.ops.quote_name(
            name='documents_documentversion_active_timestamp_max'
        ),
        timestamp=schema_editor.connection.ops.quote_name(
            name='timestamp'
        ),
        timestamp_max=schema_editor.connection.ops.quote_name(
            name='timestamp_max'
        )
    )

    cursor_main.execute(query_main)

    FETCH_SIZE = 100000

    quoted_documents_document = schema_editor.connection.ops.quote_name(
        name='documents_document'
    )
    quoted_id = schema_editor.connection.ops.quote_name(name='id')
    quoted_version_active_id = schema_editor.connection.ops.quote_name(
        name='version_active_id'
    )

    while True:
        rows = cursor_main.fetchmany(FETCH_SIZE)
        for row in rows:
            query = f'''
                UPDATE {quoted_documents_document}
                    SET {quoted_version_active_id} = {row[1]}
                WHERE {quoted_id} = {row[0]};
            '''
            cursor_document_version_active_update.execute(query)

        if not rows:
            break


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0087_document_version_active')
    ]

    operations = [
        migrations.RunPython(
            code=code_document_version_active_populate,
            reverse_code=migrations.RunPython.noop
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.platforms.platform_templates import (
    PlatformTemplate, PlatformTemplateSupervisord, Variable
)
from mayan.apps.platforms.utils import load_env_file
from mayan.apps.task_manager.classes import Worker
from mayan.settings.literals import (
    DEFAULT_DATABASE_NAME, DEFAULT_DATABASE_PASSWORD, DEFAULT_DATABASE_USER,
    DEFAULT_ELASTICSEARCH_PASSWORD, DEFAULT_RABBITMQ_CONSUMER_TIMEOUT,
    DEFAULT_RABBITMQ_PASSWORD, DEFAULT_RABBITMQ_USER, DEFAULT_RABBITMQ_VHOST,
    DEFAULT_REDIS_PASSWORD, DOCKER_ELASTIC_IMAGE_NAME,
    DOCKER_ELASTIC_IMAGE_TAG, DOCKER_IMAGE_MAYAN_NAME, DOCKER_IMAGE_MAYAN_TAG,
    DOCKER_LINUX_IMAGE_NAME, DOCKER_LINUX_IMAGE_TAG,
    DOCKER_POSTGRESQL_IMAGE_NAME, DOCKER_POSTGRESQL_IMAGE_TAG,
    DOCKER_POSTGRESQL_MAX_CONNECTIONS, DOCKER_RABBITMQ_IMAGE_NAME,
    DOCKER_RABBITMQ_IMAGE_TAG, DOCKER_REDIS_IMAGE_NAME,
    DOCKER_REDIS_IMAGE_TAG, DOCKER_TRAEFIK_IMAGE_NAME,
    DOCKER_TRAEFIK_IMAGE_TAG, LINUX_PACKAGES_DEBIAN_BASE,
    LINUX_PACKAGES_DEBIAN_BUILD, LINUX_PACKAGES_DEBIAN_MYSQL,
    LINUX_PACKAGES_DEBIAN_POSTGRESQL, LINUX_PACKAGES_DEBIAN_PYTHON
)


class PlatformTemplateDockerEntrypoint(PlatformTemplate):
    label = _(message='Template for entrypoint.sh file inside a Docker image.')
    name = 'docker_entrypoint'
    template_name = 'platforms_docker/entrypoint.tmpl'

    def get_context(self):
        context = load_env_file()
        context.update(
            {
                'SUPERVISOR_AUTORESTART': 'false',
                'workers': Worker.all()
            }
        )
        return context


class PlatformTemplateDockerComposefile(PlatformTemplate):
    label = _(message='Template that generates the Docker Compose file.')
    name = 'docker_docker_compose'
    template_name = 'platforms_docker/docker-compose.yml.tmpl'

    def __init__(self):
        self.variables = (
            Variable(
                name='DEFAULT_DATABASE_NAME',
                default=DEFAULT_DATABASE_NAME,
                environment_name='MAYAN_DEFAULT_DATABASE_NAME'
            ),
            Variable(
                name='DEFAULT_DATABASE_PASSWORD',
                default=DEFAULT_DATABASE_PASSWORD,
                environment_name='MAYAN_DEFAULT_DATABASE_PASSWORD'
            ),
            Variable(
                name='DEFAULT_DATABASE_USER',
                default=DEFAULT_DATABASE_USER,
                environment_name='MAYAN_DEFAULT_DATABASE_USER'
            ),
            Variable(
                name='DEFAULT_ELASTICSEARCH_PASSWORD',
                default=DEFAULT_ELASTICSEARCH_PASSWORD,
                environment_name='MAYAN_DEFAULT_ELASTICSEARCH_PASSWORD'
            ),
            Variable(
                name='DEFAULT_RABBITMQ_CONSUMER_TIMEOUT',
                default=DEFAULT_RABBITMQ_CONSUMER_TIMEOUT,
                environment_name='MAYAN_DEFAULT_RABBITMQ_CONSUMER_TIMEOUT'
            ),
            Variable(
                name='DEFAULT_RABBITMQ_PASSWORD',
                default=DEFAULT_RABBITMQ_PASSWORD,
                environment_name='MAYAN_DEFAULT_RABBITMQ_PASSWORD'
            ),
            Variable(
                name='DEFAULT_RABBITMQ_USER',
                default=DEFAULT_RABBITMQ_USER,
                environment_name='MAYAN_DEFAULT_RABBITMQ_USER'
            ),
            Variable(
                name='DEFAULT_RABBITMQ_VHOST',
                default=DEFAULT_RABBITMQ_VHOST,
                environment_name='MAYAN_DEFAULT_RABBITMQ_VHOST'
            ),
            Variable(
                name='DEFAULT_REDIS_PASSWORD',
                default=DEFAULT_REDIS_PASSWORD,
                environment_name='MAYAN_DEFAULT_REDIS_PASSWORD'
            ),
            Variable(
                name='DOCKER_ELASTIC_IMAGE_NAME',
                default=DOCKER_ELASTIC_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_ELASTIC_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_ELASTIC_IMAGE_TAG',
                default=DOCKER_ELASTIC_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_ELASTIC_IMAGE_TAG'
            ),
            Variable(
                name='DOCKER_IMAGE_MAYAN_NAME',
                default=DOCKER_IMAGE_MAYAN_NAME,
                environment_name='MAYAN_DOCKER_IMAGE_MAYAN_NAME'
            ),
            Variable(
                name='DOCKER_IMAGE_MAYAN_TAG',
                default=DOCKER_IMAGE_MAYAN_TAG,
                environment_name='MAYAN_DOCKER_IMAGE_MAYAN_TAG'
            ),
            Variable(
                name='DOCKER_POSTGRESQL_IMAGE_NAME',
                default=DOCKER_POSTGRESQL_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_POSTGRESQL_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_POSTGRESQL_IMAGE_TAG',
                default=DOCKER_POSTGRESQL_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_POSTGRESQL_IMAGE_TAG'
            ),
            Variable(
                name='DOCKER_POSTGRESQL_MAX_CONNECTIONS',
                default=DOCKER_POSTGRESQL_MAX_CONNECTIONS,
                environment_name='MAYAN_DOCKER_POSTGRESQL_MAX_CONNECTIONS'
            ),
            Variable(
                name='DOCKER_RABBITMQ_IMAGE_NAME',
                default=DOCKER_RABBITMQ_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_RABBITMQ_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_RABBITMQ_IMAGE_TAG',
                default=DOCKER_RABBITMQ_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_RABBITMQ_IMAGE_TAG'
            ),
            Variable(
                name='DOCKER_REDIS_IMAGE_NAME',
                default=DOCKER_REDIS_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_REDIS_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_REDIS_IMAGE_TAG',
                default=DOCKER_REDIS_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_REDIS_IMAGE_TAG'
            ),
            Variable(
                name='DOCKER_TRAEFIK_IMAGE_NAME',
                default=DOCKER_TRAEFIK_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_TRAEFIK_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_TRAEFIK_IMAGE_TAG',
                default=DOCKER_TRAEFIK_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_TRAEFIK_IMAGE_TAG'
            ),
        )


class PlatformTemplateDockerSupervisord(PlatformTemplateSupervisord):
    label = _(message='Template for Supervisord inside a Docker image.')
    name = 'docker_supervisord'
    template_name = 'platforms_docker/supervisord.tmpl'

    def get_context(self):
        return {
            'stderr_logfile': '/dev/fd/2',
            'stderr_logfile_maxbytes': '0',
            'stdout_logfile': '/dev/fd/1',
            'stdout_logfile_maxbytes': '0',
            'workers': Worker.all()
        }


class PlatformTemplateDockerfile(PlatformTemplate):
    label = _(message='Template that generates a Dockerfile file.')
    name = 'docker_dockerfile'
    template_name = 'platforms_docker/dockerfile.tmpl'

    def __init__(self):
        self.variables = (
            Variable(
                name='DOCKER_LINUX_IMAGE_NAME',
                default=DOCKER_LINUX_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_LINUX_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_LINUX_IMAGE_TAG',
                default=DOCKER_LINUX_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_LINUX_IMAGE_TAG'
            ),
        )

    def get_context(self):
        return {
            'LINUX_PACKAGES_DEBIAN_BASE': LINUX_PACKAGES_DEBIAN_BASE,
            'LINUX_PACKAGES_DEBIAN_BUILD': LINUX_PACKAGES_DEBIAN_BUILD,
            'LINUX_PACKAGES_DEBIAN_MYSQL': LINUX_PACKAGES_DEBIAN_MYSQL,
            'LINUX_PACKAGES_DEBIAN_POSTGRESQL': LINUX_PACKAGES_DEBIAN_POSTGRESQL,
            'LINUX_PACKAGES_DEBIAN_PYTHON': LINUX_PACKAGES_DEBIAN_PYTHON
        }



from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig

from .template_backends import TemplateContextEntry


class TemplatingApp(MayanAppConfig):
    app_namespace = 'templating'
    app_url = 'templating'
    has_rest_api = True
    has_static_media = True
    has_tests = True
    name = 'mayan.apps.templating'
    verbose_name = _(message='Templating')

    def ready(self):
        super().ready()

        TemplateContextEntry(
            always_available=True,
            description=_(message='Current date and time'), name='now',
            value=timezone.now
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import (
    icon_credential_backend_selection, icon_credential_delete,
    icon_credential_edit, icon_credential_list, icon_credential_setup
)
from .permissions import (
    permission_credential_create, permission_credential_delete,
    permission_credential_edit, permission_credential_view
)

link_credential_backend_selection = Link(
    icon=icon_credential_backend_selection,
    permission=permission_credential_create,
    text=_(message='Create credential'),
    view='credentials:stored_credential_backend_selection'
)
link_credential_delete = Link(
    args='resolved_object.pk',
    icon=icon_credential_delete,
    permission=permission_credential_delete,
    tags='dangerous', text=_(message='Delete'),
    view='credentials:stored_credential_delete',
)
link_credential_edit = Link(
    args='object.pk',
    icon=icon_credential_edit,
    permission=permission_credential_edit,
    text=_(message='Edit'), view='credentials:stored_credential_edit'
)
link_credential_list = Link(
    icon=icon_credential_list, text=_(message='Credential list'),
    view='credentials:stored_credential_list'
)
link_credential_setup = Link(
    condition=factory_condition_queryset_access(
        app_label='credentials', model_name='StoredCredential',
        object_permission=permission_credential_view,
        view_permission=permission_credential_create,
    ), icon=icon_credential_setup,
    text=_(message='Credentials'), view='credentials:stored_credential_list'
)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0018_auto_20150714_2227')
    ]

    operations = [
        migrations.AlterField(
            model_name='document',
            name='language',
            field=models.CharField(
                default=b'eng', max_length=8, verbose_name='Language',
                blank=True, choices=[
                    [b'aar', b'Afar'],
                    [b'abk', b'Abkhazian'],
                    [b'ace', b'Achinese'],
                    [b'ach', b'Acoli'],
                    [b'ada', b'Adangme'],
                    [b'ady', b'Adyghe; Adygei'],
                    [b'afa', b'Afro-Asiatic languages'],
                    [b'afh', b'Afrihili'],
                    [b'afr', b'Afrikaans'],
                    [b'ain', b'Ainu'],
                    [b'aka', b'Akan'],
                    [b'akk', b'Akkadian'],
                    [b'alb', b'Albanian'],
                    [b'ale', b'Aleut'],
                    [b'alg', b'Algonquian languages'],
                    [b'alt', b'Southern Altai'],
                    [b'amh', b'Amharic'],
                    [b'ang', b'English, Old (ca. 450-1100)'],
                    [b'anp', b'Angika'],
                    [b'apa', b'Apache languages'],
                    [b'ara', b'Arabic'],
                    [
                        b'arc', b'Official Aramaic (700-300 BCE); '
                        b'Imperial Aramaic (700-300 BCE)'
                    ],
                    [b'arg', b'Aragonese'],
                    [b'arm', b'Armenian'],
                    [b'arn', b'Mapudungun; Mapuche'],
                    [b'arp', b'Arapaho'],
                    [b'art', b'Artificial languages'],
                    [b'arw', b'Arawak'],
                    [b'asm', b'Assamese'],
                    [b'ast', b'Asturian; Bable; Leonese; Asturleonese'],
                    [b'ath', b'Athapascan languages'],
                    [b'aus', b'Australian languages'],
                    [b'ava', b'Avaric'],
                    [b'ave', b'Avestan'],
                    [b'awa', b'Awadhi'],
                    [b'aym', b'Aymara'],
                    [b'aze', b'Azerbaijani'],
                    [b'bad', b'Banda languages'],
                    [b'bai', b'Bamileke languages'],
                    [b'bak', b'Bashkir'],
                    [b'bal', b'Baluchi'],
                    [b'bam', b'Bambara'],
                    [b'ban', b'Balinese'],
                    [b'baq', b'Basque'],
                    [b'bas', b'Basa'],
                    [b'bat', b'Baltic languages'],
                    [b'bej', b'Beja; Bedawiyet'],
                    [b'bel', b'Belarusian'],
                    [b'bem', b'Bemba'],
                    [b'ben', b'Bengali'],
                    [b'ber', b'Berber languages'],
                    [b'bho', b'Bhojpuri'],
                    [b'bih', b'Bihari languages'],
                    [b'bik', b'Bikol'],
                    [b'bin', b'Bini; Edo'],
                    [b'bis', b'Bislama'],
                    [b'bla', b'Siksika'],
                    [b'bnt', b'Bantu languages'],
                    [b'bos', b'Bosnian'],
                    [b'bra', b'Braj'],
                    [b'bre', b'Breton'],
                    [b'btk', b'Batak languages'],
                    [b'bua', b'Buriat'],
                    [b'bug', b'Buginese'],
                    [b'bul', b'Bulgarian'],
                    [b'bur', b'Burmese'],
                    [b'byn', b'Blin; Bilin'],
                    [b'cad', b'Caddo'],
                    [b'cai', b'Central American Indian languages'],
                    [b'car', b'Galibi Carib'],
                    [b'cat', b'Catalan; Valencian'],
                    [b'cau', b'Caucasian languages'],
                    [b'ceb', b'Cebuano'],
                    [b'cel', b'Celtic languages'],
                    [b'cha', b'Chamorro'],
                    [b'chb', b'Chibcha'],
                    [b'che', b'Chechen'],
                    [b'chg', b'Chagatai'],
                    [b'chi', b'Chinese'],
                    [b'chk', b'Chuukese'],
                    [b'chm', b'Mari'],
                    [b'chn', b'Chinook jargon'],
                    [b'cho', b'Choctaw'],
                    [b'chp', b'Chipewyan; Dene Suline'],
                    [b'chr', b'Cherokee'],
                    [
                        b'chu', b'Church Slavic; Old Slavonic; Church '
                        b'Slavonic; Old Bulgarian; Old Church Slavonic'
                    ],
                    [b'chv', b'Chuvash'],
                    [b'chy', b'Cheyenne'],
                    [b'cmc', b'Chamic languages'],
                    [b'cop', b'Coptic'],
                    [b'cor', b'Cornish'],
                    [b'cos', b'Corsican'],
                    [b'cpe', b'Creoles and pidgins, English based'],
                    [b'cpf', b'Creoles and pidgins, French-based'],
                    [b'cpp', b'Creoles and pidgins, Portuguese-based'],
                    [b'cre', b'Cree'],
                    [b'crh', b'Crimean Tatar; Crimean Turkish'],
                    [b'crp', b'Creoles and pidgins'],
                    [b'csb', b'Kashubian'],
                    [b'cus', b'Cushitic languages'],
                    [b'cze', b'Czech'],
                    [b'dak', b'Dakota'],
                    [b'dan', b'Danish'],
                    [b'dar', b'Dargwa'],
                    [b'day', b'Land Dayak languages'],
                    [b'del', b'Delaware'],
                    [b'den', b'Slave (Athapascan)'],
                    [b'dgr', b'Dogrib'],
                    [b'din', b'Dinka'],
                    [b'div', b'Divehi; Dhivehi; Maldivian'],
                    [b'doi', b'Dogri'],
                    [b'dra', b'Dravidian languages'],
                    [b'dsb', b'Lower Sorbian'],
                    [b'dua', b'Duala'],
                    [b'dum', b'Dutch, Middle (ca. 1050-1350)'],
                    [b'dut', b'Dutch; Flemish'],
                    [b'dyu', b'Dyula'],
                    [b'dzo', b'Dzongkha'],
                    [b'efi', b'Efik'],
                    [b'egy', b'Egyptian (Ancient)'],
                    [b'eka', b'Ekajuk'],
                    [b'elx', b'Elamite'],
                    [b'eng', b'English'],
                    [b'enm', b'English, Middle (1100-1500)'],
                    [b'epo', b'Esperanto'],
                    [b'est', b'Estonian'],
                    [b'ewe', b'Ewe'],
                    [b'ewo', b'Ewondo'],
                    [b'fan', b'Fang'],
                    [b'fao', b'Faroese'],
                    [b'fat', b'Fanti'],
                    [b'fij', b'Fijian'],
                    [b'fil', b'Filipino; Pilipino'],
                    [b'fin', b'Finnish'],
                    [b'fiu', b'Finno-Ugrian languages'],
                    [b'fon', b'Fon'],
                    [b'fre', b'French'],
                    [b'frm', b'French, Middle (ca. 1400-1600)'],
                    [b'fro', b'French, Old (842-ca. 1400)'],
                    [b'frr', b'Northern Frisian'],
                    [b'frs', b'Eastern Frisian'],
                    [b'fry', b'Western Frisian'],
                    [b'ful', b'Fulah'],
                    [b'fur', b'Friulian'],
                    [b'gaa', b'Ga'],
                    [b'gay', b'Gayo'],
                    [b'gba', b'Gbaya'],
                    [b'gem', b'Germanic languages'],
                    [b'geo', b'Georgian'],
                    [b'ger', b'German'],
                    [b'gez', b'Geez'],
                    [b'gil', b'Gilbertese'],
                    [b'gla', b'Gaelic; Scottish Gaelic'],
                    [b'gle', b'Irish'],
                    [b'glg', b'Galician'],
                    [b'glv', b'Manx'],
                    [b'gmh', b'German, Middle High (ca. 1050-1500)'],
                    [b'goh', b'German, Old High (ca. 750-1050)'],
                    [b'gon', b'Gondi'],
                    [b'gor', b'Gorontalo'],
                    [b'got', b'Gothic'],
                    [b'grb', b'Grebo'],
                    [b'grc', b'Greek, Ancient (to 1453)'],
                    [b'gre', b'Greek, Modern (1453-)'],
                    [b'grn', b'Guarani'],
                    [b'gsw', b'Swiss German; Alemannic; Alsatian'],
                    [b'guj', b'Gujarati'],
                    [b'gwi', b"Gwich'in"],
                    [b'hai', b'Haida'],
                    [b'hat', b'Haitian; Haitian Creole'],
                    [b'hau', b'Hausa'],
                    [b'haw', b'Hawaiian'],
                    [b'heb', b'Hebrew'],
                    [b'her', b'Herero'],
                    [b'hil', b'Hiligaynon'],
                    [b'him', b'Himachali languages; Western Pahari languages'],
                    [b'hin', b'Hindi'],
                    [b'hit', b'Hittite'],
                    [b'hmn', b'Hmong; Mong'],
                    [b'hmo', b'Hiri Motu'],
                    [b'hrv', b'Croatian'],
                    [b'hsb', b'Upper Sorbian'],
                    [b'hun', b'Hungarian'],
                    [b'hup', b'Hupa'],
                    [b'iba', b'Iban'],
                    [b'ibo', b'Igbo'],
                    [b'ice', b'Icelandic'],
                    [b'ido', b'Ido'],
                    [b'iii', b'Sichuan Yi; Nuosu'],
                    [b'ijo', b'Ijo languages'],
                    [b'iku', b'Inuktitut'],
                    [b'ile', b'Interlingue; Occidental'],
                    [b'ilo', b'Iloko'],
                    [
                        b'ina', b'Interlingua (International Auxiliary '
                        b'Language Association)'
                    ],
                    [b'inc', b'Indic languages'],
                    [b'ind', b'Indonesian'],
                    [b'ine', b'Indo-European languages'],
                    [b'inh', b'Ingush'],
                    [b'ipk', b'Inupiaq'],
                    [b'ira', b'Iranian languages'],
                    [b'iro', b'Iroquoian languages'],
                    [b'ita', b'Italian'],
                    [b'jav', b'Javanese'],
                    [b'jbo', b'Lojban'],
                    [b'jpn', b'Japanese'],
                    [b'jpr', b'Judeo-Persian'],
                    [b'jrb', b'Judeo-Arabic'],
                    [b'kaa', b'Kara-Kalpak'],
                    [b'kab', b'Kabyle'],
                    [b'kac', b'Kachin; Jingpho'],
                    [b'kal', b'Kalaallisut; Greenlandic'],
                    [b'kam', b'Kamba'],
                    [b'kan', b'Kannada'],
                    [b'kar', b'Karen languages'],
                    [b'kas', b'Kashmiri'],
                    [b'kau', b'Kanuri'],
                    [b'kaw', b'Kawi'],
                    [b'kaz', b'Kazakh'],
                    [b'kbd', b'Kabardian'],
                    [b'kha', b'Khasi'],
                    [b'khi', b'Khoisan languages'],
                    [b'khm', b'Central Khmer'],
                    [b'kho', b'Khotanese;Sakan'],
                    [b'kik', b'Kikuyu; Gikuyu'],
                    [b'kin', b'Kinyarwanda'],
                    [b'kir', b'Kirghiz; Kyrgyz'],
                    [b'kmb', b'Kimbundu'],
                    [b'kok', b'Konkani'],
                    [b'kom', b'Komi'],
                    [b'kon', b'Kongo'],
                    [b'kor', b'Korean'],
                    [b'kos', b'Kosraean'],
                    [b'kpe', b'Kpelle'],
                    [b'krc', b'Karachay-Balkar'],
                    [b'krl', b'Karelian'],
                    [b'kro', b'Kru languages'],
                    [b'kru', b'Kurukh'],
                    [b'kua', b'Kuanyama; Kwanyama'],
                    [b'kum', b'Kumyk'],
                    [b'kur', b'Kurdish'],
                    [b'kut', b'Kutenai'],
                    [b'lad', b'Ladino'],
                    [b'lah', b'Lahnda'],
                    [b'lam', b'Lamba'],
                    [b'lao', b'Lao'],
                    [b'lat', b'Latin'],
                    [b'lav', b'Latvian'],
                    [b'lez', b'Lezghian'],
                    [b'lim', b'Limburgan; Limburger; Limburgish'],
                    [b'lin', b'Lingala'],
                    [b'lit', b'Lithuanian'],
                    [b'lol', b'Mongo'],
                    [b'loz', b'Lozi'],
                    [b'ltz', b'Luxembourgish; Letzeburgesch'],
                    [b'lua', b'Luba-Lulua'],
                    [b'lub', b'Luba-Katanga'],
                    [b'lug', b'Ganda'],
                    [b'lui', b'Luiseno'],
                    [b'lun', b'Lunda'],
                    [b'luo', b'Luo (Kenya and Tanzania)'],
                    [b'lus', b'Lushai'],
                    [b'mac', b'Macedonian'],
                    [b'mad', b'Madurese'],
                    [b'mag', b'Magahi'],
                    [b'mah', b'Marshallese'],
                    [b'mai', b'Maithili'],
                    [b'mak', b'Makasar'],
                    [b'mal', b'Malayalam'],
                    [b'man', b'Mandingo'],
                    [b'mao', b'Maori'],
                    [b'map', b'Austronesian languages'],
                    [b'mar', b'Marathi'],
                    [b'mas', b'Masai'],
                    [b'may', b'Malay'],
                    [b'mdf', b'Moksha'],
                    [b'mdr', b'Mandar'],
                    [b'men', b'Mende'],
                    [b'mga', b'Irish, Middle (900-1200)'],
                    [b'mic', b"Mi'kmaq; Micmac"],
                    [b'min', b'Minangkabau'],
                    [b'mis', b'Uncoded languages'],
                    [b'mkh', b'Mon-Khmer languages'],
                    [b'mlg', b'Malagasy'],
                    [b'mlt', b'Maltese'],
                    [b'mnc', b'Manchu'],
                    [b'mni', b'Manipuri'],
                    [b'mno', b'Manobo languages'],
                    [b'moh', b'Mohawk'],
                    [b'mol', b'Moldavian; Moldovan'],
                    [b'mon', b'Mongolian'],
                    [b'mos', b'Mossi'],
                    [b'mul', b'Multiple languages'],
                    [b'mun', b'Munda languages'],
                    [b'mus', b'Creek'],
                    [b'mwl', b'Mirandese'],
                    [b'mwr', b'Marwari'],
                    [b'myn', b'Mayan languages'],
                    [b'myv', b'Erzya'],
                    [b'nah', b'Nahuatl languages'],
                    [b'nai', b'North American Indian languages'],
                    [b'nap', b'Neapolitan'],
                    [b'nau', b'Nauru'],
                    [b'nav', b'Navajo; Navaho'],
                    [b'nbl', b'Ndebele, South; South Ndebele'],
                    [b'nde', b'Ndebele, North; North Ndebele'],
                    [b'ndo', b'Ndonga'],
                    [b'nds', b'Low German; Low Saxon; German, Low; Saxon, Low'],
                    [b'nep', b'Nepali'],
                    [b'new', b'Nepal Bhasa; Newari'],
                    [b'nia', b'Nias'],
                    [b'nic', b'Niger-Kordofanian languages'],
                    [b'niu', b'Niuean'],
                    [b'nno', b'Norwegian Nynorsk; Nynorsk, Norwegian'],
                    [b'nob', 'Bokm\xe5l, Norwegian; Norwegian Bokm\xe5l'],
                    [b'nog', b'Nogai'],
                    [b'non', b'Norse, Old'],
                    [b'nor', b'Norwegian'],
                    [b'nqo', b"N'Ko"],
                    [b'nso', b'Pedi; Sepedi; Northern Sotho'],
                    [b'nub', b'Nubian languages'],
                    [
                        b'nwc', b'Classical Newari; Old Newari; '
                        b'Classical Nepal Bhasa'
                    ],
                    [b'nya', b'Chichewa; Chewa; Nyanja'],
                    [b'nym', b'Nyamwezi'],
                    [b'nyn', b'Nyankole'],
                    [b'nyo', b'Nyoro'],
                    [b'nzi', b'Nzima'],
                    [b'oci', b'Occitan (post 1500)'],
                    [b'oji', b'Ojibwa'],
                    [b'ori', b'Oriya'],
                    [b'orm', b'Oromo'],
                    [b'osa', b'Osage'],
                    [b'oss', b'Ossetian; Ossetic'],
                    [b'ota', b'Turkish, Ottoman (1500-1928)'],
                    [b'oto', b'Otomian languages'],
                    [b'paa', b'Papuan languages'],
                    [b'pag', b'Pangasinan'],
                    [b'pal', b'Pahlavi'],
                    [b'pam', b'Pampanga; Kapampangan'],
                    [b'pan', b'Panjabi; Punjabi'],
                    [b'pap', b'Papiamento'],
                    [b'pau', b'Palauan'],
                    [b'peo', b'Persian, Old (ca. 600-400 B.C.)'],
                    [b'per', b'Persian'],
                    [b'phi', b'Philippine languages'],
                    [b'phn', b'Phoenician'],
                    [b'pli', b'Pali'],
                    [b'pol', b'Polish'],
                    [b'pon', b'Pohnpeian'],
                    [b'por', b'Portuguese'],
                    [b'pra', b'Prakrit languages'],
                    [
                        b'pro', 'Proven\xe7al, Old (to 1500); Occitan, '
                        'Old (to 1500)'
                    ],
                    [b'pus', b'Pushto; Pashto'],
                    [b'qaa-qtz', b'Reserved for local use'],
                    [b'que', b'Quechua'],
                    [b'raj', b'Rajasthani'],
                    [b'rap', b'Rapanui'],
                    [b'rar', b'Rarotongan; Cook Islands Maori'],
                    [b'roa', b'Romance languages'],
                    [b'roh', b'Romansh'],
                    [b'rom', b'Romany'],
                    [b'rum', b'Romanian'],
                    [b'run', b'Rundi'],
                    [b'rup', b'Aromanian; Arumanian; Macedo-Romanian'],
                    [b'rus', b'Russian'],
                    [b'sad', b'Sandawe'],
                    [b'sag', b'Sango'],
                    [b'sah', b'Yakut'],
                    [b'sai', b'South American Indian languages'],
                    [b'sal', b'Salishan languages'],
                    [b'sam', b'Samaritan Aramaic'],
                    [b'san', b'Sanskrit'],
                    [b'sas', b'Sasak'],
                    [b'sat', b'Santali'],
                    [b'scn', b'Sicilian'],
                    [b'sco', b'Scots'],
                    [b'sel', b'Selkup'],
                    [b'sem', b'Semitic languages'],
                    [b'sga', b'Irish, Old (to 900)'],
                    [b'sgn', b'Sign Languages'],
                    [b'shn', b'Shan'],
                    [b'sid', b'Sidamo'],
                    [b'sin', b'Sinhala; Sinhalese'],
                    [b'sio', b'Siouan languages'],
                    [b'sit', b'Sino-Tibetan languages'],
                    [b'sla', b'Slavic languages'],
                    [b'slo', b'Slovak'],
                    [b'slv', b'Slovenian'],
                    [b'sma', b'Southern Sami'],
                    [b'sme', b'Northern Sami'],
                    [b'smi', b'Sami languages'],
                    [b'smj', b'Lule Sami'],
                    [b'smn', b'Inari Sami'],
                    [b'smo', b'Samoan'],
                    [b'sms', b'Skolt Sami'],
                    [b'sna', b'Shona'],
                    [b'snd', b'Sindhi'],
                    [b'snk', b'Soninke'],
                    [b'sog', b'Sogdian'],
                    [b'som', b'Somali'],
                    [b'son', b'Songhai languages'],
                    [b'sot', b'Sotho, Southern'],
                    [b'spa', b'Spanish; Castilian'],
                    [b'srd', b'Sardinian'],
                    [b'srn', b'Sranan Tongo'],
                    [b'srp', b'Serbian'],
                    [b'srr', b'Serer'],
                    [b'ssa', b'Nilo-Saharan languages'],
                    [b'ssw', b'Swati'],
                    [b'suk', b'Sukuma'],
                    [b'sun', b'Sundanese'],
                    [b'sus', b'Susu'],
                    [b'sux', b'Sumerian'],
                    [b'swa', b'Swahili'],
                    [b'swe', b'Swedish'],
                    [b'syc', b'Classical Syriac'],
                    [b'syr', b'Syriac'],
                    [b'tah', b'Tahitian'],
                    [b'tai', b'Tai languages'],
                    [b'tam', b'Tamil'],
                    [b'tat', b'Tatar'],
                    [b'tel', b'Telugu'],
                    [b'tem', b'Timne'],
                    [b'ter', b'Tereno'],
                    [b'tet', b'Tetum'],
                    [b'tgk', b'Tajik'],
                    [b'tgl', b'Tagalog'],
                    [b'tha', b'Thai'],
                    [b'tib', b'Tibetan'],
                    [b'tig', b'Tigre'],
                    [b'tir', b'Tigrinya'],
                    [b'tiv', b'Tiv'],
                    [b'tkl', b'Tokelau'],
                    [b'tlh', b'Klingon; tlhIngan-Hol'],
                    [b'tli', b'Tlingit'],
                    [b'tmh', b'Tamashek'],
                    [b'tog', b'Tonga (Nyasa)'],
                    [b'ton', b'Tonga (Tonga Islands)'],
                    [b'tpi', b'Tok Pisin'],
                    [b'tsi', b'Tsimshian'],
                    [b'tsn', b'Tswana'],
                    [b'tso', b'Tsonga'],
                    [b'tuk', b'Turkmen'],
                    [b'tum', b'Tumbuka'],
                    [b'tup', b'Tupi languages'],
                    [b'tur', b'Turkish'],
                    [b'tut', b'Altaic languages'],
                    [b'tvl', b'Tuvalu'],
                    [b'twi', b'Twi'],
                    [b'tyv', b'Tuvinian'],
                    [b'udm', b'Udmurt'],
                    [b'uga', b'Ugaritic'],
                    [b'uig', b'Uighur; Uyghur'],
                    [b'ukr', b'Ukrainian'],
                    [b'umb', b'Umbundu'],
                    [b'und', b'Undetermined'],
                    [b'urd', b'Urdu'],
                    [b'uzb', b'Uzbek'],
                    [b'vai', b'Vai'],
                    [b'ven', b'Venda'],
                    [b'vie', b'Vietnamese'],
                    [b'vol', 'Volap\xfck'],
                    [b'vot', b'Votic'],
                    [b'wak', b'Wakashan languages'],
                    [b'wal', b'Wolaitta; Wolaytta'],
                    [b'war', b'Waray'],
                    [b'was', b'Washo'],
                    [b'wel', b'Welsh'],
                    [b'wen', b'Sorbian languages'],
                    [b'wln', b'Walloon'],
                    [b'wol', b'Wolof'],
                    [b'xal', b'Kalmyk; Oirat'],
                    [b'xho', b'Xhosa'],
                    [b'yao', b'Yao'],
                    [b'yap', b'Yapese'],
                    [b'yid', b'Yiddish'],
                    [b'yor', b'Yoruba'],
                    [b'ypk', b'Yupik languages'],
                    [b'zap', b'Zapotec'],
                    [b'zbl', b'Blissymbols; Blissymbolics; Bliss'],
                    [b'zen', b'Zenaga'],
                    [b'zgh', b'Standard Moroccan Tamazight'],
                    [b'zha', b'Zhuang; Chuang'],
                    [b'znd', b'Zande languages'],
                    [b'zul', b'Zulu'],
                    [b'zun', b'Zuni'],
                    [b'zxx', b'No linguistic content; Not applicable'],
                    [
                        b'zza', b'Zaza; Dimili; Dimli; Kirdki; Kirmanjki; '
                        b'Zazaki'
                    ]
                ]
            ),
            preserve_default=True
        )
    ]



from mayan.apps.forms import form_widgets, forms

from .models import Comment


class DocumentCommentDetailForm(forms.DetailForm):
    class Meta:
        extra_fields = (
            {'field': 'submit_date', 'widget': form_widgets.DateTimeInput},
            {'field': 'user'}
        )
        fields = ('text',)
        model = Comment



import logging

from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.template import Variable, VariableDoesNotExist
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.utils import get_related_field, resolve_attribute
from mayan.apps.databases.utils import (
    help_text_for_field_recursive, label_for_field_recursive
)
from mayan.apps.views.icons import icon_sort_down, icon_sort_up
from mayan.apps.views.literals import (
    TEXT_SORT_FIELD_PARAMETER, TEXT_SORT_FIELD_VARIABLE_NAME
)

from .class_mixins import TemplateObjectMixin
from .column_widgets import SourceColumnLinkWidget

logger = logging.getLogger(name=__name__)


class SourceColumn(TemplateObjectMixin):
    _registry = {}

    @staticmethod
    def sort(columns):
        columns.sort(key=lambda x: x.order)
        return columns

    @classmethod
    def get_column_matches(cls, source):
        if source == [] or source == ():
            # There are no objects to match and the object type is not
            # a queryset where the model can be obtained from empty values.
            # Short circuit and return empty columns.
            return ()

        columns = []

        try:
            # Try it as a queryset.
            source.query
        except AttributeError:
            try:
                # Try it as a list.
                item = source[0]
            except TypeError:
                # Neither a queryset nor a list.
                try:
                    # Try as a model instance or model.
                    model = source._meta.model
                except AttributeError:
                    # Not a model instance.

                    try:
                        super_class_list = source.__mro__[:-1]
                    except AttributeError:
                        # Is not a class.

                        # Try as subclass instance, check the class hierarchy.
                        for super_class in source.__class__.__mro__[:-1]:
                            columns.extend(
                                cls._registry.get(
                                    super_class, ()
                                )
                            )
                    else:
                        # Try as a subclass.
                        for super_class in super_class_list:
                            columns.extend(
                                cls._registry.get(
                                    super_class, ()
                                )
                            )

                    return columns
                else:
                    # Get model columns.
                    columns.extend(
                        cls._registry.get(
                            model, ()
                        )
                    )

                    # Get proxy columns.
                    # Remove the columns explicitly excluded.
                    # Execute after the root model columns to allow a proxy
                    # to override an existing column.
                    proxy_columns = cls._registry.get(
                        model._meta.proxy_for_model, ()
                    )
                    for proxy_column in proxy_columns:
                        if model not in proxy_column.excludes:
                            columns.append(proxy_column)

                    return columns
            else:
                # It is a list.
                return cls.get_column_matches(source=item)
        else:
            # It is a queryset.
            model = source.model

            return cls.get_column_matches(source=model)

    @classmethod
    def get_for_source(
        cls, source, exclude_identifier=False, names=None,
        only_identifier=False
    ):
        # Process columns as a set to avoid duplicate resolved column
        # detection code.
        columns = cls.get_column_matches(source=source)

        if exclude_identifier:
            columns = [
                column for column in columns if not column.is_identifier
            ]
        else:
            # exclude_identifier and only_identifier and mutually exclusive.
            if only_identifier:
                for column in columns:
                    if column.is_identifier:
                        return (column,)

                # There is no column with the identifier marker.
                return ()

        if names is not None:
            indexed_columns = {
                column.name: column for column in columns
            }

            return [
                indexed_columns[name] for name in names
            ]

        columns = SourceColumn.sort(columns=columns)

        return columns

    @classmethod
    def get_sortable_for_source(cls, source):
        result = []
        for column in cls.get_for_source(source=source):
            if column.is_sortable:
                result.append(column)

        return result

    def __init__(
        self, source, attribute=None, empty_value=None, func=None,
        help_text=None, html_extra_classes=None, include_label=False,
        is_attribute_absolute_url=False, is_object_absolute_url=False,
        is_identifier=False, is_sortable=False, kwargs=None, label=None,
        name=None, order=None, sort_field=None, widget=None
    ):
        """
        name: optional unique identifier for this source column for the
        specified source.
        """
        self._label = label
        self._help_text = help_text
        self.source = source
        self.attribute = attribute
        self.empty_value = empty_value
        self.excludes = set()
        self.func = func
        self.html_extra_classes = html_extra_classes or ''
        self.include_label = include_label
        self.is_attribute_absolute_url = is_attribute_absolute_url
        self.is_object_absolute_url = is_object_absolute_url
        self.is_identifier = is_identifier
        self.is_sortable = is_sortable
        self._kwargs = kwargs or {}
        self.name = name
        self.order = order or 0
        self.sort_field = sort_field
        self.widget = widget

        if self.is_attribute_absolute_url or self.is_object_absolute_url:
            if not self.widget:
                self.widget = SourceColumnLinkWidget

        self.__class__._registry.setdefault(
            source, []
        )
        self.__class__._registry[source].append(self)

        self._calculate_label()
        self._calculate_help_text()
        if self.is_sortable:
            field_name = self.sort_field or self.attribute
            try:
                get_related_field(
                    model=self.source, related_field_name=field_name
                )
            except FieldDoesNotExist as exception:
                raise ImproperlyConfigured(
                    '"{}" is not a field of "{}", cannot be used as a '
                    'sortable column.'.format(field_name, self.source)
                ) from exception

    def _calculate_help_text(self):
        if not self._help_text:
            if self.attribute:
                try:
                    attribute = resolve_attribute(
                        obj=self.source, attribute=self.attribute
                    )
                    self._help_text = getattr(attribute, 'help_text')
                except AttributeError:
                    try:
                        self._help_text = help_text_for_field_recursive(
                            model=self.source._meta.model,
                            name=self.attribute
                        )
                    except AttributeError:
                        self._help_text = None

        self.help_text = self._help_text

    def _calculate_label(self):
        if not self._label:
            if self.attribute:
                try:
                    attribute = resolve_attribute(
                        attribute=self.attribute, obj=self.source
                    )
                    self._label = getattr(attribute, 'short_description')
                except AttributeError:
                    try:
                        self._label = label_for_field_recursive(
                            model=self.source._meta.model,
                            name=self.attribute
                        )
                    except AttributeError:
                        self._label = self.attribute
            else:
                self._label = getattr(
                    self.func, 'short_description', _(
                        message='Unnamed function'
                    )
                )

        self.label = self._label

    def add_exclude(self, source):
        self.excludes.add(source)

    def do_kwargs_resolve(self, context):
        try:
            kwargs = self._kwargs(context)
        except TypeError:
            # Is not a callable.
            kwargs = self._kwargs

        return {
            key: Variable(var=value).resolve(context=context) for key, value in kwargs.items()
        }

    def get_absolute_url(self, obj):
        if self.is_object_absolute_url:
            return obj.get_absolute_url()
        elif self.is_attribute_absolute_url:
            result = resolve_attribute(
                attribute=self.attribute, kwargs=self.kwargs,
                obj=obj
            )
            if result:
                return result.get_absolute_url()

    def get_is_active_sort_field(self, context, reverse=False):
        previous_sort_fields = self.get_previous_sort_fields(context=context)
        sort_field = self.get_sort_field()

        if len(previous_sort_fields) == 1:
            if reverse:
                return '-{}'.format(sort_field) in previous_sort_fields
            else:
                return sort_field in previous_sort_fields

    def get_previous_sort_fields(self, context):
        previous_sort_fields = context.get(
            TEXT_SORT_FIELD_VARIABLE_NAME, None
        )

        if previous_sort_fields:
            previous_sort_fields = previous_sort_fields.split(',')
        else:
            previous_sort_fields = ()

        return previous_sort_fields

    def get_sort_field(self):
        if self.sort_field:
            return self.sort_field
        else:
            return self.attribute

    def get_sort_field_querystring(
        self, context, order=None, single_column=False
    ):
        request = self.get_request(context=context)

        # Get an mutable copy that can be modified.
        querystring = request.GET.copy()

        sort_field = self.get_sort_field()

        ascending = sort_field
        descending = '-{}'.format(sort_field)

        previous_sort_fields = list(
            self.get_previous_sort_fields(context=context)
        )

        if single_column:
            # Create a new list Remove all other fields from the list.
            if order == 'ascending':
                previous_sort_fields = []
            elif order == 'descending':
                previous_sort_fields = [ascending]
            elif order == 'clear':
                previous_sort_fields = [descending]
            else:
                previous_sort_fields = list(
                    set(previous_sort_fields).intersection(
                        (ascending, descending)
                    )
                )

        if ascending not in previous_sort_fields and descending not in previous_sort_fields:
            previous_sort_fields.append(ascending)
        elif descending in previous_sort_fields:
            previous_sort_fields.remove(descending)
        else:
            previous_sort_fields.insert(
                previous_sort_fields.index(ascending), descending
            )
            previous_sort_fields.remove(ascending)

        querystring[TEXT_SORT_FIELD_PARAMETER] = ','.join(previous_sort_fields)

        return '?{}'.format(
            querystring.urlencode()
        )

    def get_sort_icon(self, context):
        previous_sort_fields = self.get_previous_sort_fields(context=context)
        sort_field = self.get_sort_field()

        if sort_field in previous_sort_fields:
            return icon_sort_down
        elif '-{}'.format(sort_field) in previous_sort_fields:
            return icon_sort_up

    def resolve(self, context):
        kwargs = self.do_kwargs_resolve(context=context)

        if self.attribute:
            try:
                result = resolve_attribute(
                    attribute=self.attribute, kwargs=kwargs,
                    obj=context['object']
                )
            except Exception as exception:
                raise AttributeError(
                    'Unable to resolve SourceColumn attribute `{}` for object `{}`; {}.'.format(
                        self.attribute, context['object'], exception
                    )
                ) from exception
        elif self.func:
            result = self.func(context=context, **kwargs)
        else:
            result = context['object']

        self.absolute_url = self.get_absolute_url(
            obj=context['object']
        )
        if self.widget:
            try:
                request = self.get_request(context=context)
            except VariableDoesNotExist:
                """
                Don't attempt to render and return the value if any.
                """
            else:
                widget_instance = self.widget(
                    column=self, request=request, value=result
                )
                return widget_instance.render()

        if not result:
            if self.empty_value:
                return self.empty_value
            else:
                return result
        else:
            return result



from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import (
    permission_document_type_edit, permission_document_type_view
)
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin

from ..serializers import DocumentTypeMetadataTypeSerializer

from .mixins import MetadataTypeFilterAPIMixin


class APIDocumentTypeMetadataTypeListView(
    ExternalObjectAPIViewMixin, MetadataTypeFilterAPIMixin,
    generics.ListCreateAPIView
):
    """
    get: Returns a list of selected document type's metadata types.
    post: Add a metadata type to the selected document type.
    """
    external_object_class = DocumentType
    external_object_pk_url_kwarg = 'document_type_id'
    mayan_external_object_permission_map = {
        'GET': permission_document_type_view,
        'POST': permission_document_type_edit
    }
    serializer_class = DocumentTypeMetadataTypeSerializer

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class APIDocumentTypeMetadataTypeView(
    ExternalObjectAPIViewMixin, MetadataTypeFilterAPIMixin,
    generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Remove a metadata type from a document type.
    get: Retrieve the details of a document type metadata type.
    patch: Edit the selected document type metadata type.
    put: Edit the selected document type metadata type.
    """
    external_object_class = DocumentType
    external_object_pk_url_kwarg = 'document_type_id'
    lookup_url_kwarg = 'metadata_type_id'
    mayan_external_object_permission_map = {
        'DELETE': permission_document_type_edit,
        'GET': permission_document_type_view,
        'PATCH': permission_document_type_edit,
        'PUT': permission_document_type_edit
    }
    serializer_class = DocumentTypeMetadataTypeSerializer

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}



from rest_framework.filters import BaseFilterBackend, OrderingFilter

from mayan.apps.acls.models import AccessControlList
from mayan.apps.databases.utils import get_model_ordering_fields


class MayanObjectPermissionsFilter(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        permission = self.get_mayan_object_permission(
            request=request, view=view
        )

        if permission:
            return AccessControlList.objects.restrict_queryset(
                queryset=queryset, permission=permission,
                user=request.user
            )
        else:
            return queryset

    def get_mayan_object_permission(self, request, view):
        try:
            method_get_mayan_object_permission = getattr(
                view, 'get_mayan_object_permission_map'
            )
        except AttributeError:
            mayan_object_permission_map = getattr(
                view, 'mayan_object_permission_map', {}
            )

            permission = mayan_object_permission_map.get(
                request.method, None
            )

            return permission
        else:
            return method_get_mayan_object_permission()


class MayanSortingFilter(OrderingFilter):
    ordering_param = '_ordering'

    def get_default_valid_fields(self, queryset, view, context=None):
        return ()

    def get_valid_fields(self, queryset, view, context={}):
        view_ordering_fields = getattr(view, 'ordering_fields', None)

        if not view_ordering_fields:
            model = getattr(queryset, 'model', None)

            if model:
                view.ordering_fields = get_model_ordering_fields(model=model)

        return super().get_valid_fields(
            context=context, queryset=queryset, view=view
        )



from django.db import models
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.validators import YAMLValidator
from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import (
    EventManagerMethodAfter, EventManagerSave
)

from ..events import event_workflow_template_edited
from ..literals import FIELD_TYPE_CHOICES, WIDGET_CLASS_CHOICES

from .workflow_transition_field_model_mixins import (
    WorkflowTransitionFieldBusinessLogicMixin
)
from .workflow_transition_models import WorkflowTransition

__all__ = ('WorkflowTransitionField',)


class WorkflowTransitionField(
    ExtraDataModelMixin, WorkflowTransitionFieldBusinessLogicMixin,
    models.Model
):
    _ordering_fields = ('label', 'name', 'required', 'widget_kwargs')

    transition = models.ForeignKey(
        on_delete=models.CASCADE, related_name='fields',
        to=WorkflowTransition, verbose_name=_(message='Transition')
    )
    field_type = models.PositiveIntegerField(
        choices=FIELD_TYPE_CHOICES, verbose_name=_(message='Type')
    )
    name = models.CharField(
        help_text=_(
            message='The name that will be used to identify this field in '
            'other parts of the workflow system.'
        ), max_length=128, verbose_name=_(message='Internal name')
    )
    label = models.CharField(
        help_text=_(
            message='The field name that will be shown on the user interface.'
        ), max_length=128, verbose_name=_(message='Label')
    )
    default = lookup = models.TextField(
        blank=True, null=True, help_text=_(
            message='Optional default value for the field. Can be a template.'
        ), verbose_name=_(message='Default')
    )
    lookup = models.TextField(
        blank=True, null=True, help_text=_(
            message='Enter a template to render. Must result in a comma '
            'delimited string.'
        ), verbose_name=_(message='Lookup')
    )
    help_text = models.TextField(
        blank=True, help_text=_(
            message='An optional message that will help users better '
            'understand the purpose of the field and data to provide.'
        ), verbose_name=_(message='Help text')
    )
    required = models.BooleanField(
        default=False, help_text=_(
            message='Whether this field needs to be filled out or not to '
            'proceed.'
        ), verbose_name=_(message='Required')
    )
    widget = models.PositiveIntegerField(
        blank=True, choices=WIDGET_CLASS_CHOICES, help_text=_(
            message='An optional class to change the default presentation of '
            'the field.'
        ), null=True, verbose_name=_(message='Widget class')
    )
    widget_kwargs = models.TextField(
        blank=True, help_text=_(
            message='A group of keyword arguments to customize the widget. '
            'Use YAML format.'
        ), validators=[
            YAMLValidator()
        ], verbose_name=_(message='Widget keyword arguments')
    )

    class Meta:
        ordering = ('label',)
        unique_together = ('transition', 'name')
        verbose_name = _(message='Workflow transition field')
        verbose_name_plural = _(message='Workflow transition fields')

    def __str__(self):
        return self.label

    @method_event(
        action_object='self',
        event_manager_class=EventManagerMethodAfter,
        event=event_workflow_template_edited,
        target='transition.workflow'
    )
    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'action_object': 'self',
            'event': event_workflow_template_edited,
            'target': 'transition.workflow'
        },
        edited={
            'action_object': 'self',
            'event': event_workflow_template_edited,
            'target': 'transition.workflow'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)



from django.utils.translation import gettext_lazy as _

DEFAULT_EVENT_LIST_EXPORT_FILENAME = 'events_list.csv'

DEFAULT_EVENTS_DISABLE_ASYNCHRONOUS_MODE = False

DEFAULT_EVENTS_PRUNE_BACKEND = None
DEFAULT_EVENTS_PRUNE_BACKEND_ARGUMENTS = {}
DEFAULT_EVENTS_PRUNE_TASK_INTERVAL = 60 * 60 * 24 * 30  # 30 days

EVENT_MANAGER_ORDER_AFTER = 1
EVENT_MANAGER_ORDER_BEFORE = 2

EVENT_TYPE_NAMESPACE_NAME = 'events'
EVENT_EVENTS_CLEARED_NAME = 'event_cleared'
EVENT_EVENTS_EXPORTED_NAME = 'event_exported'

TEXT_UNKNOWN_EVENT_ID = _(message='Unknown or obsolete event type: %s')



from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.app_manager.apps import MayanAppConfig
from mayan.apps.common.menus import (
    menu_facet, menu_list_facet, menu_object, menu_return, menu_secondary,
    menu_tools
)
from mayan.apps.common.signals import (
    signal_post_initial_setup, signal_post_upgrade
)
from mayan.apps.navigation.source_columns import SourceColumn

from .handlers import (
    handler_search_backend_initialize, handler_search_backend_upgrade
)
from .links import (
    link_saved_resultset_delete_single, link_saved_resultset_list,
    link_search, link_saved_resultset_result_list, link_search_advanced,
    link_search_again, link_search_backend_reindex,
    link_search_model_detail, link_search_model_list
)
from .permissions import (
    permission_saved_resultset_delete, permission_saved_resultset_view
)
from .search_backends import SearchBackend
from .search_fields import SearchField
from .search_models import SearchModel


class DynamicSearchApp(MayanAppConfig):
    app_namespace = 'search'
    app_url = 'search'
    has_rest_api = True
    has_static_media = True
    has_tests = True
    name = 'mayan.apps.dynamic_search'
    verbose_name = _(message='Dynamic search')

    def ready(self):
        super().ready()

        SearchModel.load_modules()
        SearchBackend._enable()

        SavedResultset = self.get_model(model_name='SavedResultset')

        ModelPermission.register(
            model=SavedResultset,
            permissions=(
                permission_saved_resultset_delete,
                permission_saved_resultset_view,
            )
        )

        SourceColumn(
            attribute='timestamp', is_identifier=True, is_sortable=True,
            source=SavedResultset
        )
        SourceColumn(
            attribute='user', include_label=True, is_sortable=True,
            source=SavedResultset
        )
        SourceColumn(
            attribute='app_label', include_label=True, source=SavedResultset
        )
        SourceColumn(
            attribute='model_name', include_label=True, source=SavedResultset
        )
        SourceColumn(
            attribute='result_count', include_label=True,
            source=SavedResultset
        )
        SourceColumn(
            attribute='search_explainer_text', include_label=True,
            source=SavedResultset
        )
        SourceColumn(
            attribute='time_to_live', include_label=True,
            source=SavedResultset
        )

        # Search model

        SourceColumn(
            attribute='label', help_text=_(
                message='The underlying database model whose content is '
                'indexed for search.'
            ), label=_(message='Model'), include_label=True,
            is_identifier=True, source=SearchModel
        )
        SourceColumn(
            attribute='full_name', help_text=_(
                message='The unique name used to reference the search model.'
            ), include_label=True, source=SearchModel
        )

        # Search field

        SourceColumn(
            attribute='label', help_text=_(
                message='The underlying database field whose content is '
                'indexed for search.'
            ), include_label=True, is_identifier=True,
            label=_(message='Field'), source=SearchField
        )
        SourceColumn(
            attribute='field_name', include_label=True, source=SearchField
        )
        SourceColumn(
            attribute='get_search_field_class_label', include_label=True,
            source=SearchField
        )
        SourceColumn(
            attribute='field_class_label', help_text=_(
                message='The underlying database field type. This '
                'determines the kind of data that is stored in the '
                'database.'
            ), label=_(message='Field class'), include_label=True,
            source=SearchField
        )
        SourceColumn(
            attribute='get_help_text', label=_(message='Description'),
            include_label=True, source=SearchField
        )

        menu_facet.bind_links(
            links=(link_search, link_search_advanced),
            sources=(
                'search:search_simple', 'search:search_advanced',
                'search:search_results'
            )
        )
        menu_object.bind_links(
            links=(link_saved_resultset_delete_single,),
            sources=(SavedResultset,)
        )
        menu_list_facet.bind_links(
            links=(link_search_model_detail,),
            sources=(SearchModel,)
        )
        menu_list_facet.bind_links(
            links=(link_saved_resultset_result_list,),
            sources=(SavedResultset,)
        )
        menu_secondary.bind_links(
            links=(link_search_again,), sources=('search:search_results',)
        )
        menu_return.bind_links(
            links=(link_saved_resultset_list,), sources=(SavedResultset,)
        )
        menu_return.bind_links(
            links=(link_search_model_list,), sources=(
                SearchField, SearchModel
            )
        )
        menu_tools.bind_links(
            links=(
                link_saved_resultset_list, link_search_backend_reindex,
                link_search_model_list,
            ),
        )

        signal_post_initial_setup.connect(
            dispatch_uid='search_handler_search_backend_initialize',
            receiver=handler_search_backend_initialize
        )

        signal_post_upgrade.connect(
            dispatch_uid='search_handler_search_backend_upgrade',
            receiver=handler_search_backend_upgrade
        )



import json

from django.db.models import Model
from django.db.models.query import QuerySet

from mayan.apps.forms import form_widgets, forms


class BackendDynamicFormMetaclass(forms.ModelFormMetaclass):
    def __new__(mcs, name, bases, attrs):
        new_class = super(BackendDynamicFormMetaclass, mcs).__new__(
            mcs=mcs, name=name, bases=bases, attrs=attrs
        )

        if new_class._meta.fields:
            new_class._meta.fields += ('backend_data',)
            widgets = getattr(
                new_class._meta, 'widgets', {}
            ) or {}
            widgets['backend_data'] = form_widgets.HiddenInput
            new_class._meta.widgets = widgets

        return new_class


class FormDynamicModelBackend(
    forms.DynamicModelForm, metaclass=BackendDynamicFormMetaclass
):
    widgets = {'backend_data': form_widgets.HiddenInput}

    def __init__(self, backend_path=None, user=None, *args, **kwargs):
        self.backend_path = backend_path
        # For the filtered fields reload method.
        self.user = user

        super().__init__(*args, **kwargs)

        backend_data = self.instance.get_backend_data()

        if backend_data:
            for field_name in self.get_backend_fields():
                self.fields[field_name].initial = backend_data.get(
                    field_name, None
                )

        self.do_fields_reload()

    def clean(self):
        data = super().clean()

        # Consolidate the dynamic fields into a single JSON field called
        # 'backend_data'.
        backend_data = {}

        for field_name, field_data in self.schema['fields'].items():
            backend_data[field_name] = data.pop(
                field_name, field_data.get('default', None)
            )
            if isinstance(backend_data[field_name], QuerySet):
                # Flatten the queryset to a list of ids.
                backend_data[field_name] = list(
                    backend_data[field_name].values_list('id', flat=True)
                )
            elif isinstance(backend_data[field_name], Model):
                # Store only the ID of a model instance.
                backend_data[field_name] = backend_data[field_name].pk

        data['backend_data'] = json.dumps(obj=backend_data)

        return data

    def get_backend_fields(self):
        backend_class = self.instance.get_backend_class()
        return backend_class.get_form_fields()

    def get_field_reload_attributes(self):
        return {'user': self.user}



from django.utils.translation import gettext, gettext_lazy as _

from mayan.apps.forms import form_fields

from .literals import (
    SOURCE_UNCOMPRESS_CHOICE_ALWAYS, SOURCE_UNCOMPRESS_CHOICE_ASK,
    SOURCE_UNCOMPRESS_INTERACTIVE_CHOICES
)


class SourceBackendMixinCompressed:
    uncompress_choices = SOURCE_UNCOMPRESS_INTERACTIVE_CHOICES

    @classmethod
    def get_form_field_widgets(cls):
        widgets = super().get_form_field_widgets()

        widgets.update(
            {
                'uncompress': {
                    'class': 'django.forms.widgets.Select', 'kwargs': {
                        'attrs': {'class': 'select2'}
                    }
                }
            }
        )
        return widgets

    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields.update(
            {
                'uncompress': {
                    'label': _(message='Uncompress'),
                    'class': 'django.forms.ChoiceField',
                    'default': SOURCE_UNCOMPRESS_CHOICE_ASK,
                    'help_text': _(
                        message='Whether to expand or not compressed '
                        'archives.'
                    ),
                    'kwargs': {
                        'choices': cls.uncompress_choices
                    },
                    'required': True
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _(message='Decompression'), {
                    'fields': ('uncompress',)
                },
            ),
        )

        return fieldsets

    def get_expand(self):
        if self.kwargs['uncompress'] == SOURCE_UNCOMPRESS_CHOICE_ASK:
            return self.process_kwargs['forms']['source_form'].cleaned_data.get('expand')
        else:
            if self.kwargs['uncompress'] == SOURCE_UNCOMPRESS_CHOICE_ALWAYS:
                return True
            else:
                return False

    def get_task_extra_kwargs(self):
        results = super().get_task_extra_kwargs()

        results['expand'] = self.get_expand()

        return results

    def get_upload_form_class(self, action):
        backend_instance = self

        super_upload_form_class = super().get_upload_form_class(
            action=action
        )

        class CompressedSourceUploadForm(super_upload_form_class):
            expand = form_fields.BooleanField(
                label=_(message='Expand compressed files'), required=False,
                help_text=gettext(
                    'Upload a compressed file\'s contained files as '
                    'individual documents.'
                )
            )

            def __init__(self, *args, **kwargs):
                self.field_order = ['expand']
                super().__init__(*args, **kwargs)

                if backend_instance.kwargs['uncompress'] != SOURCE_UNCOMPRESS_CHOICE_ASK or action.name == 'document_file_upload':
                    self.fields.pop('expand')

        return CompressedSourceUploadForm



from django.forms import ModelChoiceField  # NOQA
from django.forms.fields import *  # NOQA
from django.forms.fields import ChoiceField, MultipleChoiceField
from django.forms.models import ModelMultipleChoiceField

from .field_mixins import (
    FormFieldMixinFilteredQueryset, ModelFieldMixinFilteredQuerySet
)


class FormFieldFilteredModelChoice(
    FormFieldMixinFilteredQueryset, ChoiceField
):
    """Single selection filtered model choice field."""


class FormFieldFilteredModelChoiceMultiple(
    FormFieldMixinFilteredQueryset, MultipleChoiceField
):
    """Multiple selection filtered model choice field."""


class ModelFormFieldFilteredModelMultipleChoice(
    ModelFieldMixinFilteredQuerySet, ModelMultipleChoiceField
):
    """Multiple selection filtered model choice field."""



from django.utils.translation import gettext_lazy as _

from mayan.apps.document_states.classes import WorkflowAction

from .models import Cabinet
from .permissions import (
    permission_cabinet_add_document, permission_cabinet_remove_document
)

__all__ = ('CabinetAddAction', 'CabinetRemoveAction')


class CabinetAddAction(WorkflowAction):
    form_field_widgets = {
        'cabinets': {
            'class': 'django.forms.widgets.SelectMultiple', 'kwargs': {
                'attrs': {'class': 'select2'}
            }
        }
    }
    label = _(message='Add to cabinets')
    permission = permission_cabinet_add_document

    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields.update(
            {
                'cabinets': {
                    'class': 'mayan.apps.forms.form_fields.FormFieldFilteredModelChoiceMultiple',
                    'help_text': _(
                        message='Cabinets to which the document will be added.'
                    ),
                    'kwargs': {
                        'source_model': Cabinet,
                        'permission': cls.permission
                    },
                    'label': _(message='Cabinets'),
                    'required': True
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _(message='Cabinets'), {
                    'fields': ('cabinets',)
                }
            ),
        )
        return fieldsets

    def execute(self, context):
        for cabinet in self.get_cabinets():
            cabinet._document_add(
                document=context['workflow_instance'].document
            )

    def get_cabinets(self):
        return Cabinet.objects.filter(
            pk__in=self.kwargs.get(
                'cabinets', ()
            )
        )


class CabinetRemoveAction(CabinetAddAction):
    label = _(message='Remove from cabinets')
    permission = permission_cabinet_remove_document

    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields['cabinets']['help_text'] = _(
            message='Cabinets from which the document will be removed'
        )

        return fields

    def execute(self, context):
        for cabinet in self.get_cabinets():
            cabinet._document_remove(
                document=context['workflow_instance'].document
            )



from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig


class FileMetadataOpenAIApp(MayanAppConfig):
    app_namespace = 'file_metadata_openai'
    app_url = 'file_metadata_openai'
    name = 'mayan.apps.file_metadata_openai'
    verbose_name = _(message='File metadata OpenAI')



from mayan.apps.organizations.utils import get_organization_installation_url
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ContentTypeAPIViewMixin

from rest_framework.settings import api_settings
from rest_framework.exceptions import ValidationError

from .exceptions import MailerError
from .mailing_actions import ModelMailingAction
from .models import UserMailer
from .permissions import (
    permission_mailing_profile_create, permission_mailing_profile_delete,
    permission_mailing_profile_edit, permission_mailing_profile_view
)
from .serializers import (
    MailingProfileActionSerializer, MailingProfileSerializer
)
from .tasks import task_send_object


class APIMailingProfileDetailView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected mailing profile.
    get: Return the details of the selected mailing profile.
    patch: Edit the selected mailing profile.
    put: Edit the selected mailing profile.
    """
    lookup_url_kwarg = 'mailing_profile_id'
    mayan_object_permission_map = {
        'DELETE': permission_mailing_profile_delete,
        'GET': permission_mailing_profile_view,
        'PATCH': permission_mailing_profile_edit,
        'PUT': permission_mailing_profile_edit
    }
    serializer_class = MailingProfileSerializer
    source_queryset = UserMailer.objects.all()

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class APIMailingProfileListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the mailing profiles.
    post: Create a new mailing profile.
    """
    mayan_object_permission_map = {'GET': permission_mailing_profile_view}
    mayan_view_permission_map = {'POST': permission_mailing_profile_create}
    serializer_class = MailingProfileSerializer
    source_queryset = UserMailer.objects.all()

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class APIMailObjectBaseView(
    ContentTypeAPIViewMixin, generics.ObjectActionAPIView
):
    lookup_url_kwarg = 'object_id'
    serializer_class = MailingProfileActionSerializer

    def get_as_attachment(self):
        model_mailing_action = self.get_model_mailing_action()

        as_attachment = model_mailing_action.as_attachment

        return as_attachment

    def get_mailing_action_name(self):
        raise NotImplementedError

    def get_mayan_object_permission_map(self):
        model_mailing_action = self.get_model_mailing_action()

        permission = model_mailing_action.kwargs['permission']

        return permission

    def get_model_mailing_action(self):
        action_name = self.get_mailing_action_name()
        content_type = self.get_content_type()
        model = content_type.model_class()

        try:
            model_mailing_action = ModelMailingAction.get_action_for_model(
                action_name=action_name, model=model
            )
        except MailerError as exception:
            raise ValidationError(
                {
                    api_settings.NON_FIELD_ERRORS_KEY: [str(exception)]
                }, code='invalid'
            )
        else:
            return model_mailing_action

    def get_source_queryset(self):
        model_mailing_action = self.get_model_mailing_action()
        manager_name = model_mailing_action.kwargs['manager_name']
        model = self.get_content_type().model_class()

        manager = getattr(model, manager_name)

        return manager.all()

    def object_action(self, obj, request, serializer):
        as_attachment = self.get_as_attachment()
        mailing_profile = serializer.validated_data['mailing_profile']
        body = serializer.validated_data.get('body')
        content_type = self.get_content_type()
        model = content_type.model_class()
        object_name = model._meta.verbose_name
        organization_installation_url = get_organization_installation_url(
            request=self.request
        )
        recipient = serializer.validated_data['recipient']
        subject = serializer.validated_data.get('subject')

        kwargs = {
            'as_attachment': as_attachment,
            'body': body,
            'content_type_id': content_type.pk,
            'object_id': obj.pk,
            'object_name': object_name,
            'organization_installation_url': organization_installation_url,
            'recipient': recipient,
            'sender': self.request.user.email,
            'subject': subject,
            'mailing_profile_id': mailing_profile.pk,
            'user_id': self.request.user.pk
        }

        task_send_object.apply_async(kwargs=kwargs)


class APIMailObjectAttachmentView(APIMailObjectBaseView):
    """
    post: Send an object as attachment via email.
    """
    def get_mailing_action_name(self):
        return 'attachment'


class APIMailObjectLinkView(APIMailObjectBaseView):
    """
    post: Send an object link via email.
    """
    def get_mailing_action_name(self):
        return 'link'



class DocumentMetadataBusinessLogicMixin:
    @property
    def is_required(self):
        """
        Return a boolean value of True of this metadata instance's parent type
        is required for the stored document type.
        """
        return self.metadata_type.get_required_for(
            document_type=self.document.document_type
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig


class BackendsApp(MayanAppConfig):
    app_namespace = 'backends'
    app_url = 'backends'
    name = 'mayan.apps.backends'
    verbose_name = _(message='Backends')



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link

from .icons import (
    icon_download_file_delete, icon_download_file_download,
    icon_download_file_list
)

link_download_file_delete = Link(
    args='resolved_object.pk', icon=icon_download_file_delete,
    tags='dangerous', text=_(message='Delete'), view='storage:download_file_delete'
)
link_download_file_download = Link(
    args='resolved_object.pk', icon=icon_download_file_download,
    tags='new_window', text=_(message='Download'),
    view='storage:download_file_download'
)
link_download_file_list = Link(
    icon=icon_download_file_list, text=_(message='Download files'),
    view='storage:download_file_list'
)



from django.apps import apps


from .literals import STORAGE_NAME_SOURCE_CACHE_FOLDER
from .settings import setting_source_cache_maximum_size


def handler_create_source_cache(sender, **kwargs):
    Cache = apps.get_model(app_label='file_caching', model_name='Cache')

    Cache.objects.update_or_create(
        defaults={
            'maximum_size': setting_source_cache_maximum_size.value,
        }, defined_storage_name=STORAGE_NAME_SOURCE_CACHE_FOLDER
    )


def handler_delete_interval_source_periodic_task(sender, instance, **kwargs):
    Source = apps.get_model(
        app_label='sources', model_name='Source'
    )

    for source in Source.objects.all():
        backend_instance = source.get_backend_instance()

        if backend_instance.kwargs.get('document_type') == instance:
            try:
                backend_instance.delete_periodic_task()
            except AttributeError:
                """
                The source has a document type but is not a periodic source,
                """


def handler_initialize_periodic_tasks(sender, **kwargs):
    Source = apps.get_model(
        app_label='sources', model_name='Source'
    )

    for source in Source.objects.filter(enabled=True):
        backend_instance = source.get_backend_instance()
        backend_instance.update()



from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig


class MIMETypesApp(MayanAppConfig):
    name = 'mayan.apps.mime_types'
    has_tests = True
    verbose_name = _(message='MIME types')

    def ready(self, *args, **kwargs):
        super().ready(*args, **kwargs)



from mayan.apps.icons.icons import Icon

icon_announcement_create = icon_tag_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='bullhorn',
    secondary_symbol='plus'
)
icon_announcement_delete_multiple = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_announcement_delete_single = icon_announcement_delete_multiple
icon_announcement_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')
icon_announcement_list = Icon(driver_name='fontawesome', symbol='bullhorn')



from django.db import migrations
from django.utils.timezone import now


def code_duplicated_document_old_copy(apps, schema_editor):
    # Remove duplicated entries.
    # `DuplicatedDocument` allowed duplicates, `DuplicateBackendEntry` does
    # not.
    cursor_primary = schema_editor.connection.cursor()
    cursor_secondary = schema_editor.connection.cursor()
    cursor_tertiary = schema_editor.connection.cursor()

    query = '''
        SELECT DISTINCT
            {documents_duplicateddocumentold}.{document_id},
            {documents_duplicateddocumentold_documents}.{document_id}
        FROM {documents_duplicateddocumentold}
        LEFT OUTER JOIN
            {documents_duplicateddocumentold_documents} ON (
                {documents_duplicateddocumentold}.{id} = {documents_duplicateddocumentold_documents}.{duplicateddocumentold_id}
            )
        ORDER BY {documents_duplicateddocumentold}.{document_id} ASC
    '''.format(
        documents_duplicateddocumentold=schema_editor.connection.ops.quote_name(
            name='documents_duplicateddocumentold'
        ),
        documents_duplicateddocumentold_documents=schema_editor.connection.ops.quote_name(
            name='documents_duplicateddocumentold_documents'
        ),
        document_id=schema_editor.connection.ops.quote_name(
            name='document_id'
        ),
        duplicateddocumentold_id=schema_editor.connection.ops.quote_name(
            name='duplicateddocumentold_id'
        ),
        id=schema_editor.connection.ops.quote_name(name='id')
    )
    cursor_primary.execute(query)

    document_insert_query = '''
        INSERT INTO duplicates_duplicateddocument_documents (
            duplicateddocument_id,document_id
        ) VALUES {};
    '''

    last_document_id = None
    document_list = []
    duplicated_document_insert_query = '''
        INSERT INTO duplicates_duplicateddocument (
            document_id,datetime_added
        ) VALUES (%s,%s);
    '''
    duplicated_document_select_query = '''
        SELECT {duplicates_duplicateddocument}.{id} FROM
            {duplicates_duplicateddocument}
        WHERE
            {duplicates_duplicateddocument}.{document_id} = %s;
    '''.format(
        document_id=schema_editor.connection.ops.quote_name(
            name='document_id'
        ),
        duplicates_duplicateddocument=schema_editor.connection.ops.quote_name(
            name='duplicates_duplicateddocument'
        ),
        id=schema_editor.connection.ops.quote_name(name='id')
    )

    now_text = now()

    for row in cursor_primary.fetchall():
        if last_document_id != row[0]:
            cursor_tertiary.execute(
                duplicated_document_insert_query, (
                    row[0], now_text
                )
            )
            cursor_tertiary.execute(
                duplicated_document_select_query, (
                    row[0],
                )
            )
            new_instance_pk = cursor_tertiary.fetchone()[0]

            if document_list:
                final_query = document_insert_query.format(
                    (
                        '(%s,%s),' * int(
                            len(document_list) / 2
                        )
                    )[:-1]
                )
                cursor_secondary.execute(
                    final_query, document_list
                )

            document_list = []
            last_document_id = row[0]
        else:
            document_list.extend(
                (
                    new_instance_pk, row[1]
                )
            )

    if document_list:
        final_query = document_insert_query.format(
            (
                '(%s,%s),' * int(
                    len(document_list) / 2
                )
            )[:-1]
        )
        cursor_secondary.execute(
            final_query, document_list
        )

    query = '''
        DELETE FROM {documents_duplicateddocumentold_documents};
    '''.format(
        documents_duplicateddocumentold_documents=schema_editor.connection.ops.quote_name(
            name='documents_duplicateddocumentold_documents'
        )
    )
    cursor_secondary.execute(query)

    query = '''
        DELETE FROM {documents_duplicateddocumentold};
    '''.format(
        documents_duplicateddocumentold=schema_editor.connection.ops.quote_name(
            name='documents_duplicateddocumentold'
        )
    )
    cursor_secondary.execute(query)


def code_duplicated_document_old_copy_reverse(apps, schema_editor):
    DuplicatedDocumentOld = apps.get_model(
        app_label='documents', model_name='DuplicatedDocumentOld'
    )

    cursor_primary = schema_editor.connection.cursor()
    cursor_secondary = schema_editor.connection.cursor()
    query = '''
        SELECT
            {duplicates_duplicateddocument}.{id},
            {duplicates_duplicateddocument}.{datetime_added},
            {duplicates_duplicateddocument}.{document_id}
        FROM {duplicates_duplicateddocument}
    '''.format(
        datetime_added=schema_editor.connection.ops.quote_name(
            name='datetime_added'
        ),
        document_id=schema_editor.connection.ops.quote_name(
            name='document_id'
        ),
        duplicates_duplicateddocument=schema_editor.connection.ops.quote_name(
            name='duplicates_duplicateddocument'
        ),
        id=schema_editor.connection.ops.quote_name(name='id')
    )
    cursor_primary.execute(query)

    for row in cursor_primary.fetchall():
        new_instance = DuplicatedDocumentOld.objects.create(
            document_id=row[2],
            datetime_added=row[1]
        )

        query = '''
            SELECT DISTINCT
                {duplicates_duplicateddocument_documents}.{document_id}
            FROM
                {duplicates_duplicateddocument_documents}
            WHERE
                {duplicates_duplicateddocument_documents}.{duplicateddocument_id} = %s
        '''.format(
            document_id=schema_editor.connection.ops.quote_name(
                name='document_id'
            ),
            duplicateddocument_id=schema_editor.connection.ops.quote_name(
                name='duplicateddocument_id'
            ),
            duplicates_duplicateddocument_documents=schema_editor.connection.ops.quote_name(
                name='duplicates_duplicateddocument_documents'
            )
        )
        cursor_secondary.execute(
            query, (
                row[0],
            )
        )
        results = cursor_secondary.fetchall()

        if results:
            insert_query = '''
                INSERT INTO documents_duplicateddocumentold_documents (
                    duplicateddocumentold_id,document_id
                ) VALUES {};
            '''
            insert_query_final = insert_query.format(
                ','.join(
                    ['(%s,%s)'] * len(results)
                )
            )
            data = []
            for result in results:
                data.append(new_instance.pk)
                data.append(
                    result[0]
                )

            cursor_secondary.execute(
                insert_query_final, data
            )

    query = '''
        DELETE FROM {duplicates_duplicateddocument_documents};
    '''.format(
        duplicates_duplicateddocument_documents=schema_editor.connection.ops.quote_name(
            name='duplicates_duplicateddocument_documents'
        )
    )
    cursor_secondary.execute(query)

    query = '''
        DELETE FROM {duplicates_duplicateddocument};
    '''.format(
        duplicates_duplicateddocument=schema_editor.connection.ops.quote_name(
            name='duplicates_duplicateddocument'
        )
    )

    cursor_secondary.execute(query)


class Migration(migrations.Migration):
    dependencies = [
        ('duplicates', '0001_initial')
    ]

    operations = [
        migrations.RunPython(
            code=code_duplicated_document_old_copy,
            reverse_code=code_duplicated_document_old_copy_reverse
        )
    ]

    run_before = [
        ('documents', '0075_delete_duplicateddocumentold')
    ]



from mayan.apps.forms import form_widgets


class SignatureCapturesAppWidget(form_widgets.TextInput):
    template_name = 'signature_captures/widget_signature_capture.html'

    def __init__(self, attrs=None):
        attrs = attrs or {}
        attrs['class'] = 'signature-captures-capture-data'
        super().__init__(attrs=attrs)



from django.urls import re_path

from .api_views import APIContentTypeDetailView, APIContentTypeListView
from .views import (
    AboutView, FaviconRedirectView, HomeView, LicenseView, ObjectCopyView,
    RootView, SetupListView, ToolsListView
)

urlpatterns_misc = [
    re_path(
        route=r'^favicon\.ico$', view=FaviconRedirectView.as_view()
    ),
    re_path(
        route=r'^object/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/copy/$',
        name='object_copy', view=ObjectCopyView.as_view()
    )
]

urlpatterns = [
    re_path(
        route=r'^$', name='root', view=RootView.as_view()
    ),
    re_path(
        route=r'^home/$', name='home', view=HomeView.as_view()
    ),
    re_path(
        route=r'^about/$', name='about_view', view=AboutView.as_view()
    ),
    re_path(
        route=r'^license/$', name='license_view', view=LicenseView.as_view()
    ),
    re_path(
        route=r'^setup/$', name='setup_list', view=SetupListView.as_view()
    ),
    re_path(
        route=r'^tools/$', name='tools_list', view=ToolsListView.as_view()
    )
]

urlpatterns.extend(urlpatterns_misc)

passthru_urlpatterns = []

api_urls = [
    re_path(
        route=r'^content_types/$', view=APIContentTypeListView.as_view(),
        name='content_type-list'
    ),
    re_path(
        route=r'^content_types/(?P<content_type_id>[0-9]+)/$',
        name='content_type-detail', view=APIContentTypeDetailView.as_view()
    )
]



from django.apps import apps
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.acls.permissions import (
    permission_acl_edit, permission_acl_view
)
from mayan.apps.app_manager.apps import MayanAppConfig
from mayan.apps.common.classes import ModelCopy
from mayan.apps.common.menus import (
    menu_list_facet, menu_object, menu_related, menu_return, menu_secondary,
    menu_setup
)
from mayan.apps.documents.links.document_type_links import (
    link_document_type_list
)
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.forms import column_widgets
from mayan.apps.navigation.source_columns import SourceColumn
from mayan.apps.rest_api.fields import DynamicSerializerField

from .events import event_smart_link_edited
from .links import (
    link_document_smart_link_instance_list, link_document_type_smart_links,
    link_smart_link_condition_create, link_smart_link_condition_delete,
    link_smart_link_condition_edit, link_smart_link_condition_list,
    link_smart_link_create, link_smart_link_delete,
    link_smart_link_document_types, link_smart_link_edit,
    link_smart_link_instance_view, link_smart_link_list, link_smart_link_setup
)
from .methods import (
    method_document_type_smart_links_add,
    method_document_type_smart_links_remove
)
from .permissions import (
    permission_resolved_smart_link_view, permission_smart_link_delete,
    permission_smart_link_edit, permission_smart_link_view
)


class LinkingApp(MayanAppConfig):
    app_namespace = 'linking'
    app_url = 'smart_links'
    has_rest_api = True
    has_tests = True
    name = 'mayan.apps.linking'
    verbose_name = _(message='Linking')

    def ready(self):
        super().ready()

        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentType = apps.get_model(
            app_label='documents', model_name='DocumentType'
        )

        ResolvedSmartLink = self.get_model(model_name='ResolvedSmartLink')
        SmartLink = self.get_model(model_name='SmartLink')
        SmartLinkCondition = self.get_model(model_name='SmartLinkCondition')

        DocumentType.add_to_class(
            name='smart_links_add',
            value=method_document_type_smart_links_add
        )
        DocumentType.add_to_class(
            name='smart_links_remove',
            value=method_document_type_smart_links_remove
        )

        DynamicSerializerField.add_serializer(
            klass=SmartLink,
            serializer_class='mayan.apps.linking.serializers.SmartLinkSerializer'
        )

        EventModelRegistry.register(model=SmartLink)
        EventModelRegistry.register(model=SmartLinkCondition)

        ModelCopy(
            model=SmartLinkCondition
        ).add_fields(
            field_names=(
                'enabled', 'expression', 'foreign_document_data',
                'inclusion', 'negated', 'operator', 'smart_link'
            )
        )
        ModelCopy(
            model=SmartLink, bind_link=True, register_permission=True
        ).add_fields(
            field_names=(
                'conditions', 'dynamic_label', 'document_types', 'enabled',
                'label'
            )
        )

        ModelEventType.register(
            event_types=(event_smart_link_edited,), model=SmartLink
        )

        ModelPermission.register(
            model=Document, permissions=(
                permission_resolved_smart_link_view,
            )
        )
        ModelPermission.register(
            model=SmartLink, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_resolved_smart_link_view,
                permission_smart_link_delete, permission_smart_link_edit,
                permission_smart_link_view
            )
        )
        ModelPermission.register_inheritance(
            model=SmartLinkCondition, related='smart_link',
        )

        # ResolvedSmartLink

        SourceColumn(
            attribute='get_label_for', is_identifier=True,
            kwargs={'document': 'document'}, label=_(message='Label'),
            source=ResolvedSmartLink
        )

        # SmartLink

        source_column_smart_link_label = SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=SmartLink
        )
        source_column_smart_link_label.add_exclude(source=ResolvedSmartLink)
        source_column_smart_link_dynamic_label = SourceColumn(
            attribute='dynamic_label', include_label=True, is_sortable=True,
            source=SmartLink
        )
        source_column_smart_link_enabled = SourceColumn(
            attribute='enabled', include_label=True, is_sortable=True,
            source=SmartLink, widget=column_widgets.TwoStateWidget
        )

        # ResolvedSmartLink
        source_column_smart_link_dynamic_label.add_exclude(
            source=ResolvedSmartLink
        )
        source_column_smart_link_enabled.add_exclude(
            source=ResolvedSmartLink
        )

        # SmartLinkCondition

        SourceColumn(
            attribute='get_full_label', is_identifier=True,
            source=SmartLinkCondition
        )
        SourceColumn(
            attribute='enabled', include_label=True, is_sortable=True,
            source=SmartLinkCondition, widget=column_widgets.TwoStateWidget
        )

        # Document

        menu_list_facet.bind_links(
            links=(link_document_smart_link_instance_list,),
            sources=(Document,)
        )

        # Document type

        menu_list_facet.bind_links(
            links=(link_document_type_smart_links,), sources=(DocumentType,)
        )
        menu_related.bind_links(
            links=(link_smart_link_list,),
            sources=(
                DocumentType, 'documents:document_type_list',
                'documents:document_type_create'
            )
        )

        # Resolved smart link

        menu_object.bind_links(
            links=(link_smart_link_instance_view,),
            sources=(ResolvedSmartLink,)
        )

        # Smart link

        menu_list_facet.bind_links(
            exclude=(ResolvedSmartLink,),
            links=(
                link_smart_link_document_types,
                link_smart_link_condition_list
            ), sources=(SmartLink,)
        )
        menu_object.bind_links(
            exclude=(ResolvedSmartLink,),
            links=(
                link_smart_link_delete, link_smart_link_edit
            ), sources=(SmartLink,)
        )
        menu_related.bind_links(
            links=(link_document_type_list,),
            sources=(
                SmartLink, 'linking:smart_link_list',
                'linking:smart_link_create'
            )
        )
        menu_return.bind_links(
            links=(link_smart_link_list,),
            sources=(
                SmartLink, 'linking:smart_link_list',
                'linking:smart_link_create'
            )
        )
        menu_secondary.bind_links(
            links=(link_smart_link_create,),
            sources=(
                SmartLink, 'linking:smart_link_list',
                'linking:smart_link_create'
            )
        )

        # Smart link condition

        menu_object.bind_links(
            links=(
                link_smart_link_condition_edit,
                link_smart_link_condition_delete
            ), sources=(SmartLinkCondition,)
        )
        menu_secondary.bind_links(
            links=(link_smart_link_condition_create,),
            sources=(
                'linking:smart_link_condition_list',
                'linking:smart_link_condition_create',
                'linking:smart_link_condition_edit',
                'linking:smart_link_condition_delete'
            )
        )

        # Setup

        menu_setup.bind_links(
            links=(link_smart_link_setup,)
        )



from datetime import timedelta

from dateutil.parser import ParserError, isoparse, parse

from django.template import Library, TemplateSyntaxError

register = Library()


@register.filter(name='date_parse_iso')
def filter_date_parse_iso(date_string):
    """
    Takes an ISO-8601 string and converts it into a datetime object.
    """

    try:
        result = isoparse(str_in=date_string)
    except ValueError as exception:
        raise TemplateSyntaxError(
            str(exception)
        ) from exception
    else:
        return result


@register.simple_tag(name='date_parse')
def tag_date_parse(
    date_string, dayfirst=None, fuzzy=None, fuzzy_with_tokens=None,
    ignoretz=None, yearfirst=None
):
    """
    Takes a string and converts it into a datetime object.
    """

    try:
        result = parse(
            dayfirst=dayfirst, fuzzy=fuzzy,
            fuzzy_with_tokens=fuzzy_with_tokens, ignoretz=ignoretz,
            timestr=date_string, yearfirst=yearfirst
        )
    except ParserError as exception:
        raise TemplateSyntaxError(
            str(exception)
        ) from exception
    else:
        return result


@register.simple_tag(name='timedelta')
def tag_timedelta(date, **kwargs):
    """
    Takes a datetime object and applies a timedelta.
    """

    return date + timedelta(**kwargs)



from mayan.apps.rest_api import generics

from ..permissions import (
    permission_workflow_template_edit, permission_workflow_template_view
)
from ..serializers.workflow_template_transition_serializers import (
    WorkflowTemplateTransitionSerializer
)

from .api_view_mixins import ParentObjectWorkflowTemplateAPIViewMixin


class APIWorkflowTemplateTransitionListView(
    ParentObjectWorkflowTemplateAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns a list of all the workflow template transitions.
    post: Create a new workflow template transition.
    """
    mayan_external_object_permission_map = {
        'GET': permission_workflow_template_view,
        'POST': permission_workflow_template_edit
    }
    serializer_class = WorkflowTemplateTransitionSerializer

    def get_instance_extra_data(self):
        # This method is only called during POST, therefore filter only by
        # edit permission.
        return {
            '_event_actor': self.request.user,
            'workflow': self.get_workflow_template()
        }

    def get_source_queryset(self):
        return self.get_workflow_template().transitions.all()


class APIWorkflowTemplateTransitionDetailView(
    ParentObjectWorkflowTemplateAPIViewMixin,
    generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected workflow template transition.
    get: Return the details of the selected workflow template transition.
    patch: Edit the selected workflow template transition.
    put: Edit the selected workflow template transition.
    """
    mayan_object_permission_map = {
        'DELETE': permission_workflow_template_edit,
        'GET': permission_workflow_template_view,
        'PATCH': permission_workflow_template_edit,
        'PUT': permission_workflow_template_edit
    }
    lookup_url_kwarg = 'workflow_template_transition_id'
    serializer_class = WorkflowTemplateTransitionSerializer

    def get_instance_extra_data(self):
        # This method is only called during POST, therefore filter only by
        # edit permission.
        return {
            '_event_actor': self.request.user,
            'workflow': self.get_workflow_template()
        }

    def get_source_queryset(self):
        return self.get_workflow_template().transitions.all()



from django.utils.translation import gettext_lazy as _

from rest_framework.exceptions import ValidationError

from mayan.apps.events.classes import EventType
from mayan.apps.events.serializers.event_type_serializers import (
    EventTypeSerializer
)
from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import MultiKwargHyperlinkedIdentityField

from ..models.workflow_transition_trigger_models import (
    WorkflowTransitionTriggerEvent
)


class WorkflowTemplateTransitionTriggerSerializer(
    serializers.HyperlinkedModelSerializer
):
    event_type = EventTypeSerializer(
        label=_(message='Event type'), read_only=True
    )
    event_type_id = serializers.CharField(
        label=_(message='Event Type ID'), source='event_type.event_type.id',
        write_only=True
    )
    url = MultiKwargHyperlinkedIdentityField(
        label=_(message='URL'), view_kwargs=(
            {
                'lookup_field': 'transition.workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
            {
                'lookup_field': 'transition_id',
                'lookup_url_kwarg': 'workflow_template_transition_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'workflow_template_transition_trigger_id'
            }
        ), view_name='rest_api:workflow-template-transition-trigger-detail'
    )
    workflow_template_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Workflow template URL'), view_kwargs=(
            {
                'lookup_field': 'transition__workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
        ), view_name='rest_api:workflow-template-detail'
    )
    workflow_transition_id = serializers.IntegerField(
        label=_(message='Workflow transition ID'), read_only=True,
        source='transition_id'
    )
    workflow_transition_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Workflow transition URL'), view_kwargs=(
            {
                'lookup_field': 'transition.workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
            {
                'lookup_field': 'transition_id',
                'lookup_url_kwarg': 'workflow_template_transition_id'
            }
        ), view_name='rest_api:workflow-template-transition-detail'
    )

    class Meta:
        fields = (
            'event_type', 'event_type_id', 'id', 'url',
            'workflow_template_url', 'workflow_transition_id',
            'workflow_transition_url'
        )
        model = WorkflowTransitionTriggerEvent
        read_only_fields = (
            'id', 'url', 'workflow_template_url', 'workflow_transition_id',
            'workflow_transition_url'
        )

    def create(self, validated_data):
        # Unroll nested source `event_type.event_type.id`.
        event_type = validated_data.pop('event_type', None)
        event_type = event_type.get(
            'event_type', {}
        )
        event_type_id = event_type.get('id')

        if event_type_id:
            validated_data['event_type'] = EventType.get(
                id=event_type_id
            ).get_stored_event_type()

        return super().create(validated_data=validated_data)

    def update(self, instance, validated_data):
        # Unroll nested source `event_type.event_type.id`.
        event_type = validated_data.pop('event_type', None)
        event_type = event_type.get(
            'event_type', {}
        )
        event_type_id = event_type.get('id')

        if event_type_id:
            validated_data['event_type'] = EventType.get(
                id=event_type_id
            ).get_stored_event_type()

        return super().update(
            instance=instance, validated_data=validated_data
        )

    def validate_event_type_id(self, data):
        try:
            EventType.get(id=data)
        except KeyError:
            raise ValidationError(
                message=_(
                    message='Unknown or invalid event type ID `%s`'
                ) % data
            )
        else:
            return data
